import json
import os

import fastapi
import pytest
import yaml
from fastapi import testclient

from fixieai.agents import api
from fixieai.agents import exceptions
from fixieai.agents import standalone
from fixieai.agents import token


@pytest.fixture(autouse=True)
def mock_token_verifier(mocker):
    return mocker.patch.object(
        token.VerifiedTokenClaims,
        "from_token",
        return_value=token.VerifiedTokenClaims(
            agent_id="fake agent id", is_anonymous=False, token="fake token"
        ),
    )


def mock_handle_message_single(query):
    return query.text.upper()


def mock_handle_message_stream(query) -> api.AgentResponseGenerator:
    yield api.AgentResponse(api.Message(query.text.upper()))


def mock_handle_message_stream_exception(query) -> api.AgentResponseGenerator:
    yield api.AgentResponse(api.Message(query.text.upper()))
    raise ValueError("Fail!")


@pytest.fixture
def standalone_agent_single(mocker):
    mocker.patch.dict(os.environ, {"FIXIE_ALLOWED_AGENT_ID": "fake agent id"})
    return standalone.StandaloneAgent(
        handle_message=mock_handle_message_single, sample_queries=["sample query"]
    )


@pytest.fixture
def standalone_agent_stream(mocker):
    mocker.patch.dict(os.environ, {"FIXIE_ALLOWED_AGENT_ID": "fake agent id"})
    return standalone.StandaloneAgent(
        handle_message=mock_handle_message_stream, sample_queries=["sample query"]
    )


def test_standalone_handshake(standalone_agent_single):
    client = testclient.TestClient(standalone_agent_single.app())

    response = client.get("/", headers={"Authorization": "Bearer fixie-test-token"})
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/yaml"
    yaml_content = yaml.load(response.content, Loader=yaml.Loader)
    assert yaml_content == {"sample_queries": ["sample query"], "type": "standalone"}


# call this test with both the standalone_agent_single and standalone_agent_stream fixtures
def test_standalone_query_single(standalone_agent_single):
    _standalone_query(standalone_agent_single)


def test_standalone_query_stream_exception(mocker):
    def _gen(query):
        yield api.AgentResponse(api.Message(query.text.upper()))
        raise ValueError("Fail!")

    mocker.patch.dict(os.environ, {"FIXIE_ALLOWED_AGENT_ID": "fake agent id"})
    agent = standalone.StandaloneAgent(
        handle_message=_gen, sample_queries=["sample query"]
    )
    client = testclient.TestClient(agent.app(), raise_server_exceptions=False)

    with client.stream(
        "POST",
        "/",
        headers={"Authorization": "Bearer fixie-test-token"},
        json={"message": {"text": "sample text", "embeds": {}}},
    ) as response:
        assert response.status_code == 200
        [response1, response2] = response.iter_lines()
        assert json.loads(response1) == {
            "message": {"text": "SAMPLE TEXT", "embeds": {}},
            "error": None,
        }
        response2_dict = json.loads(response2)
        assert (
            response2_dict["message"]["text"]
            == exceptions.UNHANDLED_EXCEPTION_RESPONSE_TEXT
        )
        assert response2_dict["error"]["code"] == "ERR_AGENT_UNHANDLED_EXCEPTION"
        assert (
            response2_dict["error"]["message"]
            == "An unhandled exception occurred while processing the request."
        )
        assert "ValueError: Fail!\n" in response2_dict["error"]["details"]["traceback"]


def _standalone_query(agent: standalone.StandaloneAgent):
    client = testclient.TestClient(agent.app())

    response = client.post(
        "/",
        headers={"Authorization": "Bearer fixie-test-token"},
        json={"message": {"text": "sample text", "embeds": {}}},
    )
    assert response.status_code == 200
    json = response.json()
    assert json == {
        "message": {"text": "SAMPLE TEXT", "embeds": {}},
        "error": None,
    }


def test_invalid_token(standalone_agent_single, mock_token_verifier):
    mock_token_verifier.return_value = None

    fast_api = fastapi.FastAPI()
    fast_api.include_router(standalone_agent_single.api_router())
    client = testclient.TestClient(fast_api, raise_server_exceptions=False)
    headers = {"Authorization": "Bearer fixie-test-token"}
    response = client.post("/", json={"message": {"text": "Howdy"}}, headers=headers)
    assert response.status_code == 403



import os

import pytest
import yaml
from fastapi import testclient
from pydantic import dataclasses as pydantic_dataclasses

import fixieai
from fixieai import agents
from fixieai.agents import agent_base
from fixieai.agents import exceptions
from fixieai.agents import token

agent_id = "dummy"
CORPUS_RESPONSE = agents.CorpusResponse(
    partitions=[agents.CorpusPartition("newPartition")],
    page=agents.CorpusPage([agents.CorpusDocument("doc1", "doc1Content".encode())]),
)
CORPUS_RESPONSE_JSON = {
    "partitions": [
        {
            "partition": "newPartition",
            "first_page_token": None,
        }
    ],
    "page": {
        "documents": [
            {
                "source_name": "doc1",
                "content": "ZG9jMUNvbnRlbnQ=",
                "mime_type": "text/plain",
                "encoding": "UTF-8",
            }
        ],
        "next_page_token": None,
    },
}
CORPUS_REQUEST_JSON = {
    "partition": "defaultPartition",
    "page_token": "token",
}


@pytest.fixture(autouse=True)
def mock_token_verifier(mocker):
    return mocker.patch.object(
        token.VerifiedTokenClaims,
        "from_token",
        return_value=token.VerifiedTokenClaims(
            agent_id="fake agent id", is_anonymous=False, token="fake token"
        ),
    )


@pytest.fixture
def dummy_agent(mocker):
    mocker.patch.dict(os.environ, {"FIXIE_ALLOWED_AGENT_ID": "fake agent id"})

    @pydantic_dataclasses.dataclass
    class DummyMetadata:
        dummy_metadata: bool = True

    class DummyAgent(agents.AgentBase):
        def metadata(self):
            return DummyMetadata()

        def validate(self):
            pass

    agent = DummyAgent(
        oauth_params=fixieai.OAuthParams(
            client_id="dummy",
            client_secret="dummy",
            auth_uri="dummy",
            token_uri="dummy",
            scopes=["dummy"],
        ),
    )

    @agent.register_func
    def simple1(query: agents.Message) -> str:
        return "Simple response 1"

    @agent.register_func()
    def simple2(query: agents.Message):
        return "Simple response 2"

    @agent.register_func(func_name="custom")
    def simple3(query):
        return "Simple response custom"

    @agent.register_func()
    def unhandled_exception(query):
        raise ValueError("Func failed!")

    @agent.register_func()
    def agent_exception(query):
        raise exceptions.AgentException(
            response_message="Agent exception!",
            error_message="My exception occurred.",
            error_code="ERR_MY_ERROR_CODE",
            http_status_code=400,
            detail=1,
        )

    @agent.register_func(func_name="corpus")
    def check_api_routing(query):
        return "Success"

    @agent.register_corpus_func
    def load_corpus(request: agents.CorpusRequest) -> agents.CorpusResponse:
        return CORPUS_RESPONSE

    @agent.register_corpus_func(func_name="custom_name_corpus_load")
    def load_corpus_custom_name(request: agents.CorpusRequest):
        return CORPUS_RESPONSE

    @agent.register_corpus_func(func_name="simple1")
    def load_corpus_existing_fn_name(request):
        return CORPUS_RESPONSE

    @agent.register_corpus_func()
    def unhandled_corpus_exception(query):
        raise ValueError("Corpus func failed!")

    @agent.register_corpus_func()
    def corpus_agent_exception(query):
        raise exceptions.AgentException(
            response_message="Corpus agent exception!",
            error_message="My exception occurred.",
            error_code="ERR_MY_ERROR_CODE",
            http_status_code=400,
            detail=1,
        )

    return agent


def test_simple_agent_handshake(dummy_agent, mocker):
    client = testclient.TestClient(dummy_agent.app())

    # Use a small chunk size to exercise chunking logic
    mocker.patch.object(agent_base, "_RESPONSE_CHUNK_SIZE", 2)
    response = client.get("/", headers={"Authorization": "Bearer fixie-test-token"})
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/yaml"
    yaml_content = yaml.load(response.content, Loader=yaml.Loader)
    assert yaml_content == {"dummy_metadata": True}


def test_simple_agent_func_calls(dummy_agent, mock_token_verifier):
    client = testclient.TestClient(dummy_agent.app(), raise_server_exceptions=False)
    headers = {"Authorization": "Bearer fixie-test-token"}

    # Test Func[simple1]
    response = client.post(
        "/simple1", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 200
    json = response.json()
    assert json == {
        "message": {"text": "Simple response 1", "embeds": {}},
        "error": None,
    }
    mock_token_verifier.assert_called_once_with(
        "fixie-test-token", dummy_agent._jwks_client, dummy_agent._allowed_agent_id
    )

    # Test Func[simple2]
    response = client.post(
        "/simple2", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 200
    json = response.json()
    assert json == {
        "message": {"text": "Simple response 2", "embeds": {}},
        "error": None,
    }

    # Test Func[custom]
    response = client.post(
        "/custom", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 200
    json = response.json()
    assert json == {
        "message": {"text": "Simple response custom", "embeds": {}},
        "error": None,
    }

    # Test Func[corpus]
    response = client.post(
        "/corpus", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 200
    json = response.json()
    assert json == {
        "message": {"text": "Success", "embeds": {}},
        "error": None,
    }

    # Test non-existing Func[] returns 404: Not Found
    response = client.post(
        "/simple3", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 404

    # Test Func[simple1] with bad arguments returns 422: Unprocessable Entity
    response = client.post(
        "/simple1", json={"message": {"ttt": "Howdy"}}, headers=headers
    )
    assert response.status_code == 422

    # Test Func[__init__] 404: Not Found
    response = client.post(
        "/__init__", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 404

    # Test Func without auth header returns 401: Unauthorized
    response = client.post("/simple1", json={"message": {"text": "Howdy"}})
    assert response.status_code == 403


def test_agent_corpus_func_calls(dummy_agent, mock_token_verifier):
    client = testclient.TestClient(dummy_agent.app(), raise_server_exceptions=False)
    headers = {"Authorization": "Bearer fixie-test-token"}
    request_body = CORPUS_REQUEST_JSON

    # Test load_corpus
    response = client.post("/corpus/load_corpus", json=request_body, headers=headers)
    assert response.status_code == 200
    json = response.json()
    assert json == CORPUS_RESPONSE_JSON
    mock_token_verifier.assert_called_once_with(
        "fixie-test-token", dummy_agent._jwks_client, dummy_agent._allowed_agent_id
    )

    # Test load_corpus_custom_name
    response = client.post(
        "/corpus/custom_name_corpus_load", json=request_body, headers=headers
    )
    assert response.status_code == 200
    json = response.json()
    assert json == CORPUS_RESPONSE_JSON

    # Test load_corpus_existing_fn_name
    response = client.post("/corpus/simple1", json=request_body, headers=headers)
    assert response.status_code == 200
    json = response.json()
    assert json == CORPUS_RESPONSE_JSON

    # Test non-existing corpus func returns 404: Not Found
    response = client.post("/corpus/non-existing", json=request_body, headers=headers)
    assert response.status_code == 404

    # Note: corpus functions will never return a 422 since there are no
    # required fields.

    # Test Func[__init__] as corpus function returns 404: Not Found
    response = client.post("/corpus/__init__", json=request_body, headers=headers)
    assert response.status_code == 404

    # Test load_corpus without auth header returns 401: Unauthorized
    response = client.post("/corpus/load_corpus", json=request_body)
    assert response.status_code == 403


def test_simple_agent_func_failure(dummy_agent, mock_token_verifier):
    client = testclient.TestClient(dummy_agent.app(), raise_server_exceptions=False)
    headers = {"Authorization": "Bearer fixie-test-token"}

    # Test Func[unhandled_exception]
    response = client.post(
        "/unhandled_exception", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 500
    json = response.json()
    assert json["message"]["text"] == exceptions.UNHANDLED_EXCEPTION_RESPONSE_TEXT
    assert "ValueError: Func failed!\n" in json["error"]["details"]["traceback"]


def test_simple_agent_func_custom_failure(dummy_agent, mock_token_verifier):
    client = testclient.TestClient(dummy_agent.app(), raise_server_exceptions=False)
    headers = {"Authorization": "Bearer fixie-test-token"}

    # Test Func[unhandled_exception]
    response = client.post(
        "/agent_exception", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 400
    json = response.json()
    assert json["message"]["text"] == "Agent exception!"
    assert json["error"]["details"]["detail"] == 1


def test_corpus_func_failure(dummy_agent, mock_token_verifier):
    client = testclient.TestClient(dummy_agent.app(), raise_server_exceptions=False)
    headers = {"Authorization": "Bearer fixie-test-token"}

    # Test unhandled_corpus_exception
    response = client.post(
        "/corpus/unhandled_corpus_exception", json=CORPUS_REQUEST_JSON, headers=headers
    )
    assert response.status_code == 500
    json = response.json()
    assert json["message"]["text"] == exceptions.UNHANDLED_EXCEPTION_RESPONSE_TEXT
    assert "ValueError: Corpus func failed!\n" in json["error"]["details"]["traceback"]


def test_corpus_agent_exception(dummy_agent, mock_token_verifier):
    client = testclient.TestClient(dummy_agent.app(), raise_server_exceptions=False)
    headers = {"Authorization": "Bearer fixie-test-token"}

    # Test corpus_agent_exception
    response = client.post(
        "/corpus/corpus_agent_exception", json=CORPUS_REQUEST_JSON, headers=headers
    )
    assert response.status_code == 400
    json = response.json()
    assert json["message"]["text"] == "Corpus agent exception!"
    assert json["error"]["details"]["detail"] == 1


def test_registering_func_with_custom_name(dummy_agent):
    dummy_agent.register_func(lambda query: "1", func_name="name1")
    dummy_agent.register_func(lambda query: "2", func_name="name2")
    dummy_agent.register_func(lambda query: "3", func_name="name3")

    def invoke_func(name: str):
        responses = dummy_agent._funcs[name](
            fixieai.AgentQuery(fixieai.Message("test")), "agent-id"
        )
        return next(iter(responses)).message.text

    assert invoke_func("name1") == "1"
    assert invoke_func("name2") == "2"
    assert invoke_func("name3") == "3"


def test_invalid_token(dummy_agent, mock_token_verifier):
    mock_token_verifier.return_value = None

    client = testclient.TestClient(dummy_agent.app(), raise_server_exceptions=False)
    headers = {"Authorization": "Bearer fixie-test-token"}
    response = client.post(
        "/simple1", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 403


def test_invalid_token_corpus_request(dummy_agent, mock_token_verifier):
    mock_token_verifier.return_value = None

    client = testclient.TestClient(dummy_agent.app(), raise_server_exceptions=False)
    headers = {"Authorization": "Bearer fixie-test-token"}
    response = client.post(
        "/corpus/load_corpus", json=CORPUS_REQUEST_JSON, headers=headers
    )
    assert response.status_code == 403



import json
import subprocess


class Tunnel:
    """Start an Internet-accessible tunnel to the Agent running at the given host:port.

    Returns a generator that yields public URLs on which the Agent can be contacted.

    Shuts down the tunnel when the context manager exits.
    """

    def __init__(self, port: int):
        self._port = port

    def __enter__(self):
        self._proc = subprocess.Popen(
            [
                "ssh",
                "-R",
                # N.B. 127.0.0.1 must be used on Windows (not localhost or 0.0.0.0)
                f"80:127.0.0.1:{self._port}",
                "-o",
                # Need to send keepalives to prevent the connection from getting chopped
                # (see https://localhost.run/docs/faq#my-connection-is-unstable-tunnels-go-down-often)
                "ServerAliveInterval=59",
                "-o",
                "StrictHostKeyChecking=accept-new",
                "nokey@localhost.run",
                "--",
                "--output=json",
            ],
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            encoding="utf-8",
        )

        def _gen():
            while True:
                assert self._proc.stdout is not None
                line = self._proc.stdout.readline()
                if not line:
                    break
                try:
                    parsed = json.loads(line)
                    if "address" in parsed:
                        yield f"https://{parsed['address']}"
                except:
                    pass

        return _gen()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._proc.terminate()
        self._proc.wait()



#!/usr/bin/env python3

from __future__ import annotations

import logging
from typing import Any, BinaryIO, Dict, List, Optional

import requests
from gql import Client
from gql import gql
from gql.transport.requests import RequestsHTTPTransport

from fixieai import constants
from fixieai.client import utils
from fixieai.client.agent import Agent
from fixieai.client.session import Session

_CLIENT: Optional["FixieClient"] = None
_SESSION: Optional[Session] = None


def get_client() -> FixieClient:
    """Return the global FixieClient instance."""
    global _CLIENT
    if not _CLIENT:
        _CLIENT = FixieClient()
    assert _CLIENT is not None
    return _CLIENT


def get_session() -> Session:
    """Return the global Fixie Session instance."""
    global _SESSION
    if not _SESSION:
        _SESSION = Session(get_client())
    assert _SESSION is not None
    return _SESSION


def get_agents() -> Dict[str, Dict[str, str]]:
    """Return metadata about all Fixie Agents. The keys of the returned
    dictionary are Agent IDs, and the values are dictionaries containing
    metadata about each Agent."""
    return get_client().get_agents()


def query(text: str) -> str:
    """Run a query."""
    return get_session().query(text)


def get_embeds() -> List[Dict[str, Any]]:
    """Return a list of Embeds."""
    return get_session().get_embeds()


class FixieClient:
    """FixieClient is a client to the Fixie system.

    Args:
        api_key: The API key for the Fixie API server. If not provided, the
            FIXIE_API_KEY environment variable will be used. If that is not
            set, the authenticated user API key will be used, or a ValueError
            will be raised if the user is not authenticated.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        self._api_key = api_key or constants.fixie_api_key()
        logging.info(f"Using Fixie API URL: {constants.FIXIE_API_URL}")
        self._request_headers = {"Authorization": f"Bearer {self._api_key}"}
        transport = RequestsHTTPTransport(
            url=constants.FIXIE_GRAPHQL_URL,
            headers=self._request_headers,
        )
        self._gqlclient = Client(transport=transport, fetch_schema_from_transport=False)

    @property
    def gqlclient(self) -> Client:
        """Return the underlying GraphQL client used by this FixieClient."""
        return self._gqlclient

    @property
    def url(self) -> str:
        """Return the URL of the Fixie API server."""
        return constants.FIXIE_API_URL

    def clone(self) -> "FixieClient":
        """Return a new FixieClient instance with the same configuration."""
        return FixieClient(api_key=self._api_key)

    def get_agents(self) -> Dict[str, Dict[str, str]]:
        """Return metadata about all running Fixie Agents. The keys of the returned
        dictionary are the Agent handles, and the values are dictionaries containing
        metadata about each Agent."""

        query = gql(
            """
            query getAgents {
                allAgents {
                    agentId
                    name
                    description
                    moreInfoUrl
                }
            }
        """
        )
        result = self._gqlclient.execute(query)
        assert "allAgents" in result and isinstance(result["allAgents"], list)
        agents = result["allAgents"]
        return {agent["agentId"]: agent for agent in agents}

    def get_agent(self, agent_id: str) -> Agent:
        """Return an existing Agent object."""
        return Agent(self, agent_id)

    def get_agent_page_url(self, agent_id: str) -> str:
        """Return the URL of the Agent page on the Fixie Platform."""
        return f"{constants.FIXIE_AGENT_URL}/{agent_id}"

    def create_agent(
        self,
        handle: str,
        name: str,
        description: str,
        query_url: Optional[str] = None,
        func_url: Optional[str] = None,
        more_info_url: Optional[str] = None,
        published: Optional[bool] = None,
    ) -> Agent:
        """Create a new Agent.

        Args:
            handle: The handle for the new Agent. This must be unique across all
                Agents owned by this user.
            name: The name of the new Agent.
            description: A description of the new Agent.
            query_url: (deprecated) Unused.
            func_url: (deprecated) The URL of the new Agent's func endpoint.
            more_info_url: A URL with more information about the new Agent.
            published: Whether the new Agent should be published.
        """
        agent = Agent(self, f"{self.get_current_username()}/{handle}")
        agent.create_agent(
            name, description, query_url, func_url, more_info_url, published
        )
        return agent

    def get_sessions(self) -> List[str]:
        """Return a list of all session IDs."""

        query = gql(
            """
            query getSessions {
                allSessions {
                    handle
                }
            }
        """
        )
        result = self._gqlclient.execute(query)
        assert "allSessions" in result and isinstance(result["allSessions"], list)
        sessions = result["allSessions"]
        return [session["handle"] for session in sessions]

    def create_session(self, frontend_agent_id: Optional[str] = None) -> Session:
        """Create a new Session."""
        return Session(self, frontend_agent_id=frontend_agent_id)

    def get_session(self, session_id: str) -> Session:
        """Return an existing Session object."""
        return Session(self, session_id)

    def get_current_user(self) -> Dict[str, Any]:
        """Returns the user metadata of the current user."""
        query = gql(
            """
            query getUsername {
                user {
                    username
                    firstName
                    lastName
                    email
                    organization {
                        handle
                        name
                    }
                    dailyQueryLimit
                    dailyUsedQueries
                    avatar
                }
            }
        """
        )
        result = self._gqlclient.execute(query)
        assert "user" in result and isinstance(result["user"], dict)
        return result["user"]

    def get_current_username(self) -> str:
        """Returns the username of the current user."""
        user = self.get_current_user()
        assert "username" in user and isinstance(user["username"], str)
        return user["username"]

    def refresh_agent(self, agent_handle: str):
        """Indicates that an agent's prompts should be refreshed."""
        username = self.get_current_username()
        requests.post(
            f"{constants.FIXIE_REFRESH_URL}/{username}/{agent_handle}",
            headers=self._request_headers,
        ).raise_for_status()

    def create_agent_revision(
        self,
        handle: str,
        make_current: bool,
        *,
        reindex_corpora: bool = False,
        metadata: Optional[Dict[str, str]] = None,
        external_url: Optional[str] = None,
        python_gzip_tarfile: Optional[BinaryIO] = None,
    ) -> str:
        """Creates a new Agent revision.

        Args:
            handle: The handle of the Agent. Must be owned by the current user.
            make_current: Whether the new revision should be made the current (active) revision.
            reindex_corpora: Whether to reindex all corpora for the new revision.
            metadata: Optional client-provided metadata to associate with the revision.
            external_url: The URL at which the revision is hosted, if hosted externally.
            python_gzip_tarfile: A file-like of a gzip-compressed tarfile containing the files to deploy.

        Exactly one of `external_url` and `python_gzip_tarfile` must be provided.
        """

        mutation = gql(
            """
            mutation CreateAgentRevision(
                $handle: String!,
                $metadata: [RevisionMetadataKeyValuePairInput!]!,
                $makeCurrent: Boolean!,
                $externalDeployment: ExternalDeploymentInput,
                $managedDeployment: ManagedDeploymentInput,
                $reindexCorpora: Boolean!,
            ) {
                createAgentRevision(agentHandle: $handle, makeCurrent: $makeCurrent, revision: { 
                    metadata: $metadata
                    externalDeployment: $externalDeployment
                    managedDeployment: $managedDeployment
                }, reindexCorpora: $reindexCorpora) {
                    revision {
                        id
                    }
                }
            }
            """
        )

        with utils.patched_gql_file_uploader(
            python_gzip_tarfile, "upload.tar.gz", "application/gzip"
        ):
            result = self._gqlclient.execute(
                mutation,
                variable_values={
                    "handle": handle,
                    "metadata": [
                        {"key": key, "value": value} for key, value in metadata.items()
                    ]
                    if metadata
                    else [],
                    "makeCurrent": make_current,
                    "reindexCorpora": reindex_corpora,
                    "externalDeployment": {"url": external_url}
                    if external_url
                    else None,
                    "managedDeployment": {
                        "environment": "PYTHON",
                        "codePackage": python_gzip_tarfile,
                    }
                    if python_gzip_tarfile
                    else None,
                },
                upload_files=True,
            )

        revision_id = result["createAgentRevision"]["revision"]["id"]
        assert isinstance(revision_id, str)
        return revision_id

    def get_current_agent_revision(self, handle: str) -> Optional[str]:
        """Gets the current (active) revision of an agent."""
        agent = self.get_agent(handle)
        query = gql(
            """
            query GetRevisionId($agentId: String!) {
                agentById(agentId: $agentId) {
                    currentRevision {
                        id
                    }
                }
            }
            """
        )

        response = self._gqlclient.execute(
            query,
            variable_values={"agentId": agent.agent_id},
        )

        revision = response["agentById"]["currentRevision"]
        return revision["id"] if revision is not None else None

    def set_current_agent_revision(self, handle: str, revision_id: str):
        """Sets the current (active) revision of an agent."""

        mutation = gql(
            """
            mutation SetCurrentAgentRevision(
                $handle: String!,
                $currentRevisionId: ID!) {
                updateAgent(
                    agentData: {
                        handle: $handle,
                        currentRevisionId: $currentRevisionId
                    }
                ) {
                    agent {
                        agentId
                    }
                }
            }
            """
        )

        _ = self._gqlclient.execute(
            mutation,
            variable_values={"handle": handle, "currentRevisionId": revision_id},
        )

    def delete_agent_revision(self, handle: str, revision_id: str):
        """Deletes an Agent revision."""

        mutation = gql(
            """
            mutation DeleteRevision($handle: String!, $revisionId: ID!) {
                deleteAgentRevision(agentHandle: $handle, revisionId: $revisionId) {
                    agent {
                        agentId
                    }
                }
            }
            """
        )
        _ = self._gqlclient.execute(
            mutation, variable_values={"handle": handle, "revisionId": revision_id}
        )

    def deploy_agent(
        self,
        handle: str,
        gzip_tarfile: BinaryIO,
    ):
        """Deploys an agent implementation.

        Args:
            handle: The handle of the Agent to deploy. Must be owned by the current user.
            gzip_tarfile: A file-like of a gzip-compressed tarfile containing the files to deploy.
        """
        return self.create_agent_revision(
            handle,
            make_current=True,
            python_gzip_tarfile=gzip_tarfile,
        )



import pytest

from fixieai import constants
from fixieai.cli.auth import user_config

EXAMPLE_URL = "https://example.fixie.ai"


@pytest.fixture
def config_without_environment():
    return user_config.UserConfig.from_dict(
        {
            "fixie_api_key": "default-token",
        }
    )


@pytest.fixture
def config_with_environment():
    return user_config.UserConfig.from_dict(
        {
            "fixie_api_key": "default-token",
            "environment_auth_tokens": [
                {"url": EXAMPLE_URL, "auth_token": "environment-token"}
            ],
        }
    )


def test_user_config_no_env(config_without_environment):
    assert config_without_environment.auth_token == "default-token"


def test_user_config_no_matching_env(config_with_environment):
    assert config_with_environment.auth_token == "default-token"


def test_user_config_matching_env(mocker, config_with_environment):
    mocker.patch.object(constants, "FIXIE_API_URL", EXAMPLE_URL)
    assert config_with_environment.auth_token == "environment-token"


def test_set_token_multiple_environments(mocker, config_without_environment):
    url_1, token_1 = "https://test1.fixie.ai", "token1"
    mocker.patch.object(constants, "FIXIE_API_URL", url_1)
    config_without_environment.auth_token = token_1
    assert config_without_environment.auth_token == token_1

    url_2, token_2 = "https://test2.fixie.ai", "token2"
    mocker.patch.object(constants, "FIXIE_API_URL", url_2)
    config_without_environment.auth_token = token_2
    assert config_without_environment.auth_token == token_2

    # If we switch the URL back we should get the first token
    mocker.patch.object(constants, "FIXIE_API_URL", url_1)
    assert config_without_environment.auth_token == token_1


def test_config_to_from_yaml(config_with_environment, config_without_environment):
    assert (
        user_config.UserConfig.from_yaml(config_without_environment.to_yaml())
        == config_without_environment
    )
    assert (
        user_config.UserConfig.from_yaml(config_with_environment.to_yaml())
        == config_with_environment
    )



from fixieai.agents.agent_base import AgentBase
from fixieai.agents.api import AgentQuery
from fixieai.agents.api import AgentResponse
from fixieai.agents.api import Embed
from fixieai.agents.api import Message
from fixieai.agents.code_shot import CodeShotAgent
from fixieai.agents.corpora import CorpusDocument
from fixieai.agents.corpora import CorpusPage
from fixieai.agents.corpora import CorpusPartition
from fixieai.agents.corpora import CorpusRequest
from fixieai.agents.corpora import CorpusResponse
from fixieai.agents.corpora import CustomCorpus
from fixieai.agents.corpora import DocumentCorpus
from fixieai.agents.corpora import UrlDocumentCorpus
from fixieai.agents.llm_settings import LlmSettings
from fixieai.agents.oauth import OAuthHandler
from fixieai.agents.oauth import OAuthParams
from fixieai.agents.standalone import StandaloneAgent
from fixieai.agents.user_storage import UserStorage

__all__ = [
    "AgentBase",
    "AgentQuery",
    "AgentResponse",
    "CodeShotAgent",
    "CorpusDocument",
    "CorpusPage",
    "CorpusPartition",
    "CorpusRequest",
    "CorpusResponse",
    "CustomCorpus",
    "DocumentCorpus",
    "Embed",
    "LlmSettings",
    "Message",
    "StandaloneAgent",
    "OAuthHandler",
    "OAuthParams",
    "UrlDocumentCorpus",
    "UserStorage",
]



import importlib.metadata

from fixieai.agents import AgentQuery
from fixieai.agents import AgentResponse
from fixieai.agents import CodeShotAgent
from fixieai.agents import CorpusDocument
from fixieai.agents import CorpusPage
from fixieai.agents import CorpusPartition
from fixieai.agents import CorpusRequest
from fixieai.agents import CorpusResponse
from fixieai.agents import DocumentCorpus
from fixieai.agents import Embed
from fixieai.agents import LlmSettings
from fixieai.agents import Message
from fixieai.agents import OAuthHandler
from fixieai.agents import OAuthParams
from fixieai.agents import StandaloneAgent
from fixieai.agents import UrlDocumentCorpus
from fixieai.agents import UserStorage
from fixieai.client import FixieClient
from fixieai.client import get_agents
from fixieai.client import get_client
from fixieai.client import get_embeds
from fixieai.client import query

__all__ = [
    "AgentQuery",
    "AgentResponse",
    "Embed",
    "Message",
    "CodeShotAgent",
    "CorpusDocument",
    "CorpusPage",
    "CorpusPartition",
    "CorpusRequest",
    "CorpusResponse",
    "DocumentCorpus",
    "StandaloneAgent",
    "LlmSettings",
    "OAuthParams",
    "OAuthHandler",
    "UrlDocumentCorpus",
    "UserStorage",
    "FixieClient",
    "get_agents",
    "get_client",
    "get_embeds",
    "query",
]

__version__ = importlib.metadata.version(__name__)



import json

from fixieai.agents import corpora


def test_request_serialization():
    request = corpora.CorpusRequest(partition="partition", page_token="token")
    as_dict = request.to_dict()
    assert isinstance(as_dict, dict)

    # Assert query can be reconstructed from as_dict
    request_reconstructed = corpora.CorpusRequest.from_dict(as_dict)
    assert isinstance(request_reconstructed, corpora.CorpusRequest)
    assert request_reconstructed == request

    # Assert serialize_as_dict produces a json serializable string
    as_json = json.dumps(as_dict)
    as_dict_from_json = json.loads(as_json)
    request_reconstructed = corpora.CorpusRequest.from_dict(as_dict_from_json)
    assert isinstance(request_reconstructed, corpora.CorpusRequest)
    assert request_reconstructed == request


def test_response_serialization():
    response = corpora.CorpusResponse(
        partitions=[
            corpora.CorpusPartition("part1", "part1token"),
            corpora.CorpusPartition("part2"),
        ],
        page=corpora.CorpusPage(
            [
                corpora.CorpusDocument("doc1", "content".encode()),
                corpora.CorpusDocument("doc2", "moreContent".encode("cp037"), "cp037"),
            ],
            "continuationToken",
        ),
    )
    as_dict = response.to_dict()
    assert isinstance(as_dict, dict)

    # Assert response can be reconstructed from as_dict
    response_reconstructed = corpora.CorpusResponse.from_dict(as_dict)
    assert isinstance(response_reconstructed, corpora.CorpusResponse)
    assert response_reconstructed == response

    # Assert serialize_as_dict produces a json serializable string
    as_json = json.dumps(as_dict)
    as_dict_from_json = json.loads(as_json)
    response_reconstructed = corpora.CorpusResponse.from_dict(as_dict_from_json)
    assert isinstance(response_reconstructed, corpora.CorpusResponse)
    assert response_reconstructed == response



import os

import fastapi
import pytest
import yaml
from fastapi import testclient

import fixieai
from fixieai import agents
from fixieai.agents import code_shot
from fixieai.agents import token

agent_id = "dummy"
BASE_PROMPT = "I am a simple dummy agent."
FEW_SHOTS = """
Q: Sample query 1
Ask Func[simple1]: Simple argument
Func[simple1] says: Simple response
A: Simple final response

Q: Sample query 2
Ask Func[simple2]: Simple argument
Func[simple2] says: Simple response
A: Simple final response
"""
CORPORA = [
    agents.UrlDocumentCorpus(urls=["http://example.com/doc1.txt"]),
]


@pytest.fixture(autouse=True)
def mock_token_verifier(mocker):
    return mocker.patch.object(
        token.VerifiedTokenClaims,
        "from_token",
        return_value=token.VerifiedTokenClaims(
            agent_id="fake agent id", is_anonymous=False, token="fake token"
        ),
    )


@pytest.fixture
def dummy_code_shot_agent(mocker):
    mocker.patch.dict(os.environ, {"FIXIE_ALLOWED_AGENT_ID": "fake agent id"})
    agent = agents.CodeShotAgent(
        BASE_PROMPT,
        FEW_SHOTS,
        CORPORA,
        conversational=False,
        oauth_params=fixieai.OAuthParams(
            client_id="dummy",
            client_secret="dummy",
            auth_uri="dummy",
            token_uri="dummy",
            scopes=["dummy"],
        ),
        llm_settings=agents.LlmSettings(
            model="test-model", temperature=0.42, maximum_tokens=42
        ),
    )

    @agent.register_corpus_func()
    def load_empty_custom_corpus(
        request: agents.CorpusRequest,
    ) -> agents.CorpusResponse:
        return agents.CorpusResponse()

    return agent


def test_code_shot_handshake(dummy_code_shot_agent, mocker):
    fast_api = fastapi.FastAPI()
    fast_api.include_router(dummy_code_shot_agent.api_router())
    client = testclient.TestClient(fast_api)

    response = client.get("/", headers={"Authorization": "Bearer fixie-test-token"})
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/yaml"
    yaml_content = yaml.load(response.content, Loader=yaml.Loader)
    assert yaml_content == {
        "base_prompt": dummy_code_shot_agent.base_prompt,
        "few_shots": dummy_code_shot_agent.few_shots,
        "corpora": [
            {
                "urls": ["http://example.com/doc1.txt"],
                "exclude_patterns": None,
                "auth_token_func": None,
                "loader": None,
            },
            {"func_name": "load_empty_custom_corpus"},
        ],
        "conversational": dummy_code_shot_agent.conversational,
        "response_model": {
            "model": dummy_code_shot_agent.llm_settings.model,
            "temperature": dummy_code_shot_agent.llm_settings.temperature,
            "maximum_tokens": dummy_code_shot_agent.llm_settings.maximum_tokens,
        },
        "type": "code_shot",
    }


def test_split_few_shots():
    few_shots_list = code_shot._split_few_shots(FEW_SHOTS)
    assert few_shots_list == [
        """
Q: Sample query 1
Ask Func[simple1]: Simple argument
Func[simple1] says: Simple response
A: Simple final response""",
        """Q: Sample query 2
Ask Func[simple2]: Simple argument
Func[simple2] says: Simple response
A: Simple final response""",
    ]



import contextlib
import contextvars
import functools
import logging
import os
import sys
from typing import Iterable, Optional, Tuple

import starlette.types
import wrapt  # type: ignore

import fixieai.constants

FAKE_OPENAI_KEY = "sk-fixie-openai-key"

logger = logging.getLogger(__name__)

# A ContextVar that allows the current request token to be used for proxied OpenAI requests.
_current_request_token: contextvars.ContextVar[Optional[str]] = contextvars.ContextVar(
    "openai_proxy.current_request_token",
    default=None,
)


def enable_openai_proxy() -> bool:
    """Enables the Fixie OpenAI proxy for outbound OpenAI requests made in the context of an agent request.

    Returns a value that indicates whether or not the proxy is enabled. If either OPENAI_API_BASE or
    OPENAI_API_KEY is already set, the proxy will not be enabled.

    The OpenAIProxyRequestTokenForwarderMiddleware must be used for the proxy to use valid auth tokens.
    """

    if "OPENAI_API_BASE" in os.environ or "OPENAI_API_KEY" in os.environ:
        logger.warning(
            "OpenAI proxy is not enabled; leave OPENAI_API_BASE and OPENAI_API_KEY unset to use the Fixie OpenAI proxy."
        )
        return False

    assert (
        "openai" not in sys.modules
    ), "OpenAI has already been loaded. The OpenAI proxy should be enabled before loading OpenAI."

    def _patch_api_requestor(module):
        original_init = module.APIRequestor.__init__

        @functools.wraps(original_init)
        def patched_init(self, *args, **kwargs):
            original_init(self, *args, **kwargs)

            if self.api_base == fixieai.constants.FIXIE_OPENAI_PROXY_URL:
                if self.api_key != FAKE_OPENAI_KEY:
                    logger.warning(
                        "An OpenAI key was provided for the Fixie OpenAI proxy. "
                        "Ignoring it in favor of the current Fixie request token."
                    )

                self.api_key = _current_request_token.get()
                if self.api_key is None:
                    raise ValueError(
                        "The Fixie request token was not set. "
                        "The Fixie OpenAI proxy can only be used in the context of an agent request."
                    )
            elif self.api_key == FAKE_OPENAI_KEY:
                raise ValueError(
                    "The OpenAI base URL was changed but the key was not. "
                    "Leave the OpenAI base URL unspecified to use the Fixie OpenAI proxy or provide an OpenAI api key."
                )

        module.APIRequestor.__init__ = patched_init

    wrapt.register_post_import_hook(_patch_api_requestor, "openai.api_requestor")
    os.environ["OPENAI_API_BASE"] = fixieai.constants.FIXIE_OPENAI_PROXY_URL
    os.environ["OPENAI_API_KEY"] = FAKE_OPENAI_KEY

    return True


class OpenAIProxyRequestTokenForwarderMiddleware:
    """Middleware that allows the current request token to be used for proxied OpenAI requests."""

    def __init__(self, app: starlette.types.ASGIApp):
        self.app = app

    async def __call__(
        self,
        scope: starlette.types.Scope,
        receive: starlette.types.Receive,
        send: starlette.types.Send,
    ):
        BEARER_PREFIX = b"Bearer "

        with contextlib.ExitStack() as stack:
            http_headers: Iterable[Tuple[bytes, bytes]] = (
                scope["headers"] if scope["type"] in ("http", "websocket") else {}
            )

            # If there's a Bearer token in the headers, put it in the context var.
            for header, value in http_headers:
                if header.lower() == b"authorization" and value.startswith(
                    BEARER_PREFIX
                ):
                    reset_token = _current_request_token.set(
                        value[len(BEARER_PREFIX) :].decode()
                    )
                    stack.callback(lambda: _current_request_token.reset(reset_token))
                    break

            await self.app(scope, receive, send)



import dataclasses
import json
from typing import Callable, List, Optional

import fastapi
import fastapi.security

from fixieai.agents import agent_base
from fixieai.agents import agent_func
from fixieai.agents import api
from fixieai.agents import metadata
from fixieai.agents import oauth


class StandaloneAgent(agent_base.AgentBase):
    """An agent that handles queries directly.

    To make a StandaloneAgent, pass a function with the following signature

    def handle(query: fixieai.Message) -> ReturnType:
            ...

    where ReturnType is one of `str`, `fixieai.Message`, or `fixie.AgentResponse`.
    """

    def __init__(
        self,
        handle_message: Callable,
        sample_queries: Optional[List[str]] = None,
        oauth_params: Optional[oauth.OAuthParams] = None,
    ):
        super().__init__(oauth_params)

        if isinstance(handle_message, agent_func.AgentFunc):
            self._handle_message: agent_func.AgentFunc = handle_message
        else:
            self._handle_message = agent_func.AgentQueryFunc.create(
                handle_message,
                oauth_params,
                default_message_type=api.Message,
                allow_generator=True,
            )
        self._sample_queries = sample_queries

    def metadata(self) -> metadata.Metadata:
        return metadata.StandaloneAgentMetadata(sample_queries=self._sample_queries)

    def validate(self):
        pass

    def api_router(self) -> fastapi.APIRouter:
        router = super().api_router()
        router.add_api_route("/", self._serve_query, methods=["POST"])
        return router

    def _serve_query(
        self,
        query: api.AgentQuery,
        credentials: fastapi.security.HTTPAuthorizationCredentials = fastapi.Depends(
            fastapi.security.HTTPBearer()
        ),
    ) -> fastapi.responses.Response:
        """Verifies the request is a valid request from Fixie, and dispatches it to
        the previously specified `handle_message` function. Depending on the return
        value of that function, either a single or a streaming response is returned.
        """
        token_claims = super()._check_credentials(credentials)

        output = self._handle_message(query, token_claims)
        return fastapi.responses.StreamingResponse(
            (json.dumps(dataclasses.asdict(resp)) + "\n" for resp in output),
            media_type="application/json",
        )



import contextlib
import functools
import importlib
import inspect
import os
import sys
import threading
import time
from typing import Tuple

import dotenv

import fixieai
from fixieai import agents
from fixieai.agents import agent_func
from fixieai.agents import openai_proxy
from fixieai.cli.agent import agent_config


@contextlib.contextmanager
def _ensure_serving_disabled():
    original_serve = agents.AgentBase.serve

    def _fail(*_, **__):
        raise RuntimeError(
            "agent.serve() must not be called while your agent is being imported."
        )

    agents.AgentBase.serve = _fail  # type: ignore[assignment]
    try:
        yield
    finally:
        agents.AgentBase.serve = original_serve  # type: ignore[assignment]


def load_agent_from_path(
    path: str,
) -> Tuple[agent_config.AgentConfig, agents.AgentBase]:
    """Loads an Agent and its config from a path."""

    path = agent_config.normalize_path(path)
    config = agent_config.load_config(path)
    agent_dir = os.path.dirname(path) or "."

    dotenv_path = os.path.join(agent_dir, ".env")
    if os.path.exists(dotenv_path):
        dotenv.load_dotenv(dotenv_path)

    openai_proxy.enable_openai_proxy()

    # Inject the agent directory into PYTHONPATH
    sys.path.insert(0, agent_dir)

    entry_point_parts = config.entry_point.split(":", 1)
    module_name = entry_point_parts[0]
    attr = entry_point_parts[1] if len(entry_point_parts) == 2 else "agent"

    with _ensure_serving_disabled():
        module = importlib.import_module(module_name)
    impl = getattr(module, attr)

    if isinstance(impl, agents.AgentBase):
        agent_impl = impl
    elif isinstance(impl, agent_func.AgentFunc):
        agent_impl = agents.StandaloneAgent(impl)
    elif inspect.isfunction(impl):
        func = agent_func.AgentQueryFunc.create(
            impl, oauth_params=None, default_message_type=str, allow_generator=True
        )
        agent_impl = agents.StandaloneAgent(func)
    else:
        raise ValueError("Entrypoint must refer to an agent instance or function.")

    agent_impl.validate()
    return config, agent_impl


def _refresh_agent_async(agent_handle: str):
    """Asynchronously pings Fixie to refresh the given agent_id."""

    def _refresh_agent_sync():
        # Pause a second to give the agent a moment to start up.
        time.sleep(1)
        fixieai.FixieClient().refresh_agent(agent_handle)

    thread = threading.Thread(target=_refresh_agent_sync, daemon=True)
    thread.start()


def uvicorn_app_factory():
    """Returns an app that can be used to serve an agent with uvicorn.

    If the FIXIE_REFRESH_ON_STARTUP environment variable is set to "true" an agent refresh
    will be triggered each time the agent starts up.
    """
    config, impl = load_agent_from_path(".")
    fastapi_app = impl.app()
    fastapi_app.add_middleware(openai_proxy.OpenAIProxyRequestTokenForwarderMiddleware)

    # The CLI no longer uses this, but continue to honor it to preserve compatibility with CLI 0.2.x.
    if os.getenv("FIXIE_REFRESH_ON_STARTUP") == "true":
        fastapi_app.add_event_handler(
            "startup", functools.partial(_refresh_agent_async, config.handle)
        )

    return fastapi_app


if __name__ == "__main__":
    # Load the agent (typically within an agent-specific venv) to ensure the Python module can be loaded
    # without errors. See `fixie agent serve` and `fixie agent deploy --validate`.
    load_agent_from_path(".")



from typing import List, Optional, Union

from fixieai.agents import agent_base
from fixieai.agents import corpora
from fixieai.agents import llm_settings
from fixieai.agents import metadata
from fixieai.agents import oauth
from fixieai.agents import utils


class CodeShotAgent(agent_base.AgentBase):
    """A CodeShot agent.

    To make a CodeShot agent, simply pass a BASE_PROMPT and FEW_SHOTS:

        BASE_PROMPT = "A summary of what this agent does; how it does it; and its
        personality"

        FEW_SHOTS = '''
        Q: <Sample query that this agent supports>
        A: <Desired response for this query>

        Q: <Another sample query>
        A: <Desired response for this query>
        '''

        agent = CodeShotAgent(BASE_PROMPT, FEW_SHOTS)

    You can have FEW_SHOTS as a single string of all your few-shots separated by 2 new
    lines, or as an explicit list of one few-shot per index.

    Your few-shots may reach out to other Agents in the fixie ecosystem by
    "Ask Agent[agent_id]: <query to pass>", or reach out to some python functions
    by "Ask Func[func_name]: <query to pass>".

    There are a series of default runtime `Func`s provided by the platform available for
    your agents to consume. For a full list of default runtime `Func`s, refer to:
        http://docs.fixie.ai/XXX

    You may also need to write your own python functions here to be consumed by your
    agent. To make a function accessible by an agent, you'd need to register it by
    `@agent.register_func`. Example:

        @agent.register_func
        def func_name(query: fixieai.Message) -> ReturnType:
            ...

        , where ReturnType is one of `str`, `fixieai.Message`, or `fixie.AgentResponse`.

    Note that in the above, we are using the decorator `@agent.register_func` to
    register this function with the agent instance we just created.

    To check out the default `Func`s that are provided in Fixie, see:
        http://docs.fixie.ai/XXX

    """

    def __init__(
        self,
        base_prompt: str,
        few_shots: Union[str, List[str]],
        corpora: Optional[List[corpora.UrlDocumentCorpus]] = None,
        conversational: bool = False,
        oauth_params: Optional[oauth.OAuthParams] = None,
        llm_settings: Optional[llm_settings.LlmSettings] = None,
    ):
        super().__init__(oauth_params)

        if isinstance(few_shots, str):
            few_shots = _split_few_shots(few_shots)

        self.base_prompt = base_prompt
        self.few_shots = few_shots
        self.url_corpora = corpora
        self.conversational = conversational
        self.llm_settings = llm_settings

        utils.strip_prompt_lines(self)

    @property
    def corpora(self):
        url_corpora = self.url_corpora or []
        custom_corpora = [
            corpora.CustomCorpus(func_name=f) for f in self._corpus_funcs.keys()
        ]
        return url_corpora + custom_corpora if url_corpora or custom_corpora else None

    def metadata(self) -> metadata.Metadata:
        return metadata.CodeShotAgentMetadata(
            self.base_prompt,
            self.few_shots,
            self.corpora,
            self.conversational,
            self.llm_settings,
        )

    def validate(self):
        utils.validate_code_shot_agent(self)


def _split_few_shots(few_shots: str) -> List[str]:
    """Split a long string of all few-shots into a list of few-shot strings."""
    # First, strip all lines to remove bad spaces.
    few_shots = "\n".join(line.strip() for line in few_shots.splitlines())
    # Then, split by "\n\nQ:".
    few_shot_splits = few_shots.split("\n\nQ:")
    few_shot_splits = [few_shot_splits[0]] + [
        "Q:" + few_shot for few_shot in few_shot_splits[1:]
    ]
    return few_shot_splits



from typing import List, Literal, Optional, Union

from pydantic import dataclasses as pydantic_dataclasses

from fixieai.agents import corpora as fixie_corpora
from fixieai.agents import llm_settings


@pydantic_dataclasses.dataclass
class StandaloneAgentMetadata:
    """Metadata for a Fixie Standalone Agent."""

    sample_queries: Optional[List[str]] = None
    type: Literal["standalone"] = "standalone"


@pydantic_dataclasses.dataclass
class CodeShotAgentMetadata:
    """Metadata for a Fixie CodeShot Agent."""

    base_prompt: str
    few_shots: List[str]
    corpora: Optional[
        List[Union[fixie_corpora.UrlDocumentCorpus, fixie_corpora.CustomCorpus]]
    ] = None
    conversational: bool = False
    response_model: Optional[llm_settings.LlmSettings] = None
    type: Literal["code_shot"] = "code_shot"


Metadata = Union[StandaloneAgentMetadata, CodeShotAgentMetadata]



"""This module holds objects that represent the API interface by which Agents talk to
Fixie ecosystem."""

from __future__ import annotations

import base64
import dataclasses
from typing import Dict, Generator, Optional

import requests
from pydantic import dataclasses as pydantic_dataclasses


@pydantic_dataclasses.dataclass
class Embed:
    """An Embed represents a binary object attached to a Message."""

    # The MIME content type of the object, e.g., "image/png" or "application/json".
    content_type: str

    # A public URL where the object can be downloaded. This can be a data URI.
    uri: str

    @property
    def content(self) -> bytes:
        """Retrieves the content for this Embed object."""
        if self.uri.startswith("data:"):
            return base64.b64decode(self.uri.split(",")[1])
        return requests.get(self.uri).content

    @content.setter
    def content(self, content: bytes):
        """Sets the content of the Embed object as a data URI."""
        self.uri = f"data:base64,{base64.b64encode(content).decode('utf-8')}"

    @property
    def text(self) -> str:
        """Retrieves the content of the Embed object as a string."""
        return self.content.decode("utf-8")

    @text.setter
    def text(self, text: str):
        """Sets the content of the Embed object as a string."""
        self.content = text.encode("utf-8")


@pydantic_dataclasses.dataclass
class Message:
    """A Message represents a single message sent to a Fixie agent."""

    # The text of the message.
    text: str

    # A mapping of embed keys to Embed objects.
    embeds: Dict[str, Embed] = dataclasses.field(default_factory=dict)


@pydantic_dataclasses.dataclass
class AgentQuery:
    """A standalone query sent to a Fixie agent."""

    # The contents of the query.
    message: Message

    # This is an access token associated with the user for whom this query was
    # created. Agents wishing to make queries to other agents, or to other
    # Fixie services, should carry this token in the query so that it
    # can be tied back to the original user.
    access_token: Optional[str] = None


@pydantic_dataclasses.dataclass
class AgentError:
    """An error that occurred in the process of generating a response."""

    # A code representing the error.
    code: str

    # A message describing the error that will be displayed to the user.
    message: str

    # Additional debugging context from the error.
    details: Optional[Dict]


@pydantic_dataclasses.dataclass
class AgentResponse:
    """A response message from an Agent."""

    # The text of the response message.
    message: Message

    # Error details, if an error occurred generating the response. Note that `message` should still be included,
    # but may indicate that an error occurred that prevented the agent from fully handling the request.
    error: Optional[AgentError] = None


AgentResponseGenerator = Generator[AgentResponse, None, None]



from __future__ import annotations

import abc
import collections.abc
import inspect
import typing
from typing import (
    Any,
    Callable,
    Generator,
    Generic,
    Iterable,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_type_hints,
)

from fixieai.agents import api
from fixieai.agents import corpora
from fixieai.agents import exceptions
from fixieai.agents import oauth
from fixieai.agents import token
from fixieai.agents import user_storage
from fixieai.agents.api import AgentResponse
from fixieai.agents.api import Message
from fixieai.agents.corpora import CorpusResponse

F = TypeVar("F")  # from (request)
T = TypeVar("T")  # to (response)
R = TypeVar("R")  # any supported response type

# An ArgumentMapper converts a query + token claims into a supported func argument (e.g. "query" or "user_storage")
ArgumentMapper = Callable[[F, token.VerifiedTokenClaims], Any]

# BoundArgumentMapper is an ArgumentMapper that has been bound to a named parameter.
BoundArgumentMapper = Tuple[str, ArgumentMapper]


class AgentFunc(abc.ABC, Generic[F, T, R]):
    """A Python function that can be invoked by Fixie."""

    @abc.abstractmethod
    def adapt_result(self, result: R) -> Generator[T, None, None]:
        pass

    @abc.abstractmethod
    def adapt_exception(self, e: exceptions.AgentException) -> T:
        pass

    def __init__(
        self,
        impl: Callable[..., R],
        argument_mappers: Iterable[BoundArgumentMapper],
    ):
        self._impl = impl
        self._argument_mappers = tuple(argument_mappers)

    def __call__(
        self, agent_query: F, claims: token.VerifiedTokenClaims
    ) -> Iterable[T]:
        with exceptions.AgentException.exception_remapper():
            kwargs = {}
            for name, mapper in self._argument_mappers:
                kwargs[name] = mapper(agent_query, claims)

            result = self._impl(**kwargs)

        def _gen() -> Iterable[T]:
            # Within a generator we need to catch exceptions and convert them to responses.
            try:
                with exceptions.AgentException.exception_remapper():
                    yield from self.adapt_result(result)
            except exceptions.AgentException as e:
                yield self.adapt_exception(e)

        return _gen()


class AgentQueryFunc(
    AgentFunc[
        api.AgentQuery,
        api.AgentResponse,
        Union[
            api.AgentResponse,
            api.Message,
            str,
            Iterable[Union[api.AgentResponse, api.Message, str]],
        ],
    ]
):
    """A Python function that can be invoked by Fixie in the context of a user query.

    Wrapped functions can take up to three arguments:

    1. A message or query of type: `str`, `api.Message`, or `api.AgentQuery`;
       name: "query", or "message"; or the first parameter if no other rules apply.
    2. A user storage object of type `user_storage.UserStorage` or name "user_storage".
    3. An OAuth handler of type `oauth.OAuthHandler` or name "oauth_handler".

    The semantics of each parameter are inferred by type annotation or name/position if no
    type annotations are present.

    The function may return an `api.AgentResponse`, an `api.Message`, a `str` or an
    iterable of one of the same. If an iterable is returned and allow_multiple is
    false, all values after the first will be dropped.
    """

    def __init__(
        self,
        impl: Callable[
            ..., AgentResponse | Message | str | Iterable[AgentResponse | Message | str]
        ],
        argument_mappers: Iterable[
            Tuple[str, Callable[[F, token.VerifiedTokenClaims], Any]]
        ],
        allow_multiple_responses: bool,
    ):
        super().__init__(impl, argument_mappers)
        self._allow_multiple_responses = allow_multiple_responses

    def adapt_result(
        self,
        result: AgentResponse | Message | str | Iterable[AgentResponse | Message | str],
    ) -> Generator[AgentResponse, None, None]:
        """Adapts any allowed func return type into an Iterable of AgentResponses."""

        if isinstance(result, api.AgentResponse):
            yield result
        elif isinstance(result, api.Message):
            yield api.AgentResponse(message=result)
        elif isinstance(result, str):
            yield api.AgentResponse(message=api.Message(result))
        elif isinstance(result, Iterable):
            for value in result:
                yield from self.adapt_result(value)

                if not self._allow_multiple_responses:
                    break
        else:
            raise TypeError(
                f"The func result was type {type(result)}, but must be a str, Message, AgentResponse, or iterable."
            )

    def adapt_exception(self, e: exceptions.AgentException) -> AgentResponse:
        return e.to_response()

    @classmethod
    def create(
        cls,
        func: Callable,
        oauth_params: Optional[oauth.OAuthParams],
        default_message_type: Union[Type[str], Type[api.Message], Type[api.AgentQuery]],
        allow_generator: bool,
    ) -> AgentQueryFunc:
        """Constructs an AgentQueryFunc from a Python function.

        Args:
            func: the Python function to wrap
            oauth_params: optional OAuthParams, required if the func requires an OAuthHandler
            default_message_type: the message type to default to if type annotations aren't present
            allow_generator: whether the function is allowed to return multiple responses
        """

        if not inspect.isfunction(func):
            raise TypeError(
                f"Registered function {func!r} is not a function, but a {type(func)!r}."
            )
        signature = inspect.signature(func)
        func_name = func.__name__
        params = signature.parameters

        # Validate that there are no var args (*args or **kwargs).
        if any(
            param.kind in (param.VAR_KEYWORD, param.VAR_POSITIONAL)
            for param in params.values()
        ):
            raise TypeError(
                f"Registered function {func_name} cannot accept variable args: {params!r}."
            )

        # Using the signature (parameter names and type annotations) of the provided function,
        # figure out which parameters correspond to which of the possible values (message,
        # user storage, OAuth handler).
        #
        # Because the types of each of the possible values are distinct, we prioritize type
        # annotations if present, but fall back to well-known names.
        #
        # For each of the possible values we'll try to construct a `BoundArgumentMapper`
        # that can be used at runtime to set the arguments.

        # Ensure every parameter is used exactly once.
        unbound_parameter_names = set(params.keys())

        # Try to bind a query/message mapper.
        typed_message_mappers = {
            api.AgentQuery: lambda query, claims: query,
            api.Message: lambda query, claims: query.message,
            str: lambda query, claims: query.message.text,
        }
        bound_message_mapper = _bind_argument_mapper(
            func,
            unbound_parameter_names=unbound_parameter_names,
            mappers_by_type_or_name=(
                *typed_message_mappers.items(),
                ("query", typed_message_mappers[default_message_type]),
                ("message", typed_message_mappers[default_message_type]),
            ),
        )

        # Try to bind a UserStorage mapper.
        user_storage_mapper = lambda query, claims: user_storage.UserStorage(claims)
        bound_user_storage_mapper = _bind_argument_mapper(
            func,
            unbound_parameter_names=unbound_parameter_names,
            mappers_by_type_or_name=(
                (user_storage.UserStorage, user_storage_mapper),
                ("user_storage", user_storage_mapper),
            ),
        )

        # Try to bind an OAuthHandler mapper.
        oauth_mapper = lambda query, claims: oauth_params and oauth.OAuthHandler(
            oauth_params, claims
        )
        bound_oauth_mapper = _bind_argument_mapper(
            func,
            unbound_parameter_names=unbound_parameter_names,
            mappers_by_type_or_name=(
                (oauth.OAuthHandler, oauth_mapper),
                ("oauth_handler", oauth_mapper),
            ),
        )
        if bound_oauth_mapper is not None and oauth_params is None:
            raise TypeError(
                f"Function {func_name} that accepts 'oauth_handler' as an argument cannot "
                f"be registered with an agent that hasn't set 'oauth_params' in its "
                "constructor."
            )

        # If we didn't get a message handler and there's a single untyped argument
        # remaining and it's the first parameter, treat it as a message.
        #
        # def f(x): allowed, x is a message
        # def f(x, user_storage): allowed, x is a message
        # def f(user_storage, x): not allowed
        # def f(x, y): not allowed
        # def f(x: int): not allowed
        type_hints = get_type_hints(func)
        if bound_message_mapper is None and len(unbound_parameter_names) == 1:
            remaining_parameter_name = next(iter(unbound_parameter_names))
            first_parameter_name = next(iter(params.keys()))
            if (
                first_parameter_name == remaining_parameter_name
                and remaining_parameter_name not in type_hints
            ):
                bound_message_mapper = (
                    remaining_parameter_name,
                    typed_message_mappers[default_message_type],
                )
                unbound_parameter_names.remove(remaining_parameter_name)

        # All parameters should be accounted for.
        if len(unbound_parameter_names) > 0:
            raise TypeError(
                f"Registered function {func_name} had unknown parameters: {unbound_parameter_names}"
            )

        # Validate that the return type annotation is compatible.
        return_type = get_type_hints(func).get("return")
        if return_type is not None:
            VALID_RETURN_TYPES = (api.AgentResponse, api.Message, str)
            is_return_type_valid = return_type in VALID_RETURN_TYPES

            if allow_generator and not is_return_type_valid:
                # If allow_generator is set, the function can return any Iterable
                # of an allowed type.
                origin_type = typing.get_origin(return_type)
                if origin_type is not None and issubclass(
                    origin_type, collections.abc.Iterable
                ):
                    args = typing.get_args(return_type)
                    is_return_type_valid = (args == ()) or args[0] in VALID_RETURN_TYPES

            if not is_return_type_valid:
                raise TypeError(
                    f"Expected registered function to return an AgentResponse, a Message, "
                    f"or str but it returns {return_type}."
                )

        return AgentQueryFunc(
            func,
            [
                bound_mapper
                for bound_mapper in (
                    bound_message_mapper,
                    bound_user_storage_mapper,
                    bound_oauth_mapper,
                )
                if bound_mapper is not None
            ],
            allow_generator,
        )


class AgentCorpusFunc(
    AgentFunc[
        corpora.CorpusRequest,
        corpora.CorpusResponse,
        corpora.CorpusResponse,
    ]
):
    """A Python function that can be invoked by Fixie to load a custom corpus.

    Wrapped functions must take one `corpora.CorpusRequest` argument (or no
    arguments) and return a `corpora.CorpusResponse`.
    """

    def adapt_result(
        self, result: CorpusResponse
    ) -> Generator[CorpusResponse, None, None]:
        if isinstance(result, CorpusResponse):
            yield result
        else:
            raise TypeError(
                f"The func result was type {type(result)}, but must be a CorpusResponse."
            )

    def adapt_exception(self, e: exceptions.AgentException) -> CorpusResponse:
        raise e

    @classmethod
    def create(
        cls,
        func: Callable[[corpora.CorpusRequest], corpora.CorpusResponse],
    ) -> AgentCorpusFunc:
        """Constructs an AgentCorpusFunc from a Python function.

        Args:
            func: the Python function to wrap
        """

        if not inspect.isfunction(func):
            raise TypeError(
                f"Registered function {func!r} is not a function, but a {type(func)!r}."
            )
        signature = inspect.signature(func)
        func_name = func.__name__
        params = signature.parameters

        # Validate that there are no var args (*args or **kwargs).
        if any(
            param.kind in (param.VAR_KEYWORD, param.VAR_POSITIONAL)
            for param in params.values()
        ):
            raise TypeError(
                f"Registered function {func_name} cannot accept variable args: {params!r}."
            )

        if len(params) > 1:
            raise TypeError(
                f"Registered function {func_name} must have one or zero arguments."
            )
        param = next(iter(params.keys()), None)
        if param:
            type_hints = get_type_hints(func)
            if param in type_hints and type_hints.get(param) != corpora.CorpusRequest:
                raise TypeError(
                    f"Registered function {func_name}'s parameter has the wrong type: {param}. (Should be CorpusRequest.)"
                )
            bound_request_mapper = (param, lambda request, claims: request)
        else:
            bound_request_mapper = None

        # Validate that the return type annotation is compatible.
        return_type = get_type_hints(func).get("return")
        if return_type is not None and return_type != corpora.CorpusResponse:
            raise TypeError(
                f"Expected registered function to return a CorpusResponse, but it returns {return_type}"
            )

        return AgentCorpusFunc(
            func, [bound_request_mapper] if bound_request_mapper else []
        )


def _bind_argument_mapper(
    func: Callable,
    unbound_parameter_names: Set[str],
    mappers_by_type_or_name: Iterable[Tuple[Union[type, str], ArgumentMapper]],
) -> Optional[BoundArgumentMapper]:
    """Attempts to select an argument name to which a mapper can be bound.

    If a mapper is bound, the name will be removed from `unbound_parameter_names`.
    """

    type_hints = get_type_hints(func)

    for type_or_name, mapper in mappers_by_type_or_name:
        for name in unbound_parameter_names:
            if isinstance(type_or_name, type) and type_hints.get(name) == type_or_name:
                unbound_parameter_names.remove(name)
                return name, mapper
            elif (
                name not in type_hints
                and isinstance(type_or_name, str)
                and name == type_or_name
            ):
                unbound_parameter_names.remove(name)
                return name, mapper

    return None



from typing import Optional

import click

from fixieai import FixieClient
from fixieai.cli.session import console


def validate_agent_exists(ctx, param, agent_id):
    if agent_id is not None:
        client = ctx.obj.client
        agent = client.get_agent(agent_id)
        if not agent.valid:
            raise click.BadParameter(f"Agent {agent_id} does not exist")
    return agent_id


@click.group(help="Session-related commands.")
def session():
    pass


@session.command("list", help="Lists sessions.")
@click.pass_context
def list_sessions(ctx):
    client = ctx.obj.client
    session_ids = client.get_sessions()
    for session_id in session_ids:
        click.secho(f"{session_id}", fg="green")


def web_option(func):
    return click.option(
        "--web",
        is_flag=True,
        help="Open the session in the web interface.",
    )(func)


def run_session(
    client: FixieClient,
    session_id: Optional[str] = None,
    agent: Optional[str] = None,
    message: Optional[str] = None,
    use_console: bool = False,
    web: bool = False,
    keep: bool = False,
    verbose: bool = False,
):
    """Runs a session.

    Args:
        client: The FixieClient instance.
        session_id: The ID of the session to use. If not provided, a new session is created.
        agent: The agent to use for the session.
        message: The message to send to the session.
        use_console: Whether to run the session in the console.
        web: Whether to open the session in the web interface.
        keep: Whether to keep the session after it is finished.
        verbose: Whether to print verbose output.
    """
    if session_id is not None:
        session = client.get_session(session_id)
    else:
        session = client.create_session(agent)
    if verbose:
        click.secho(f"Session ID: ", nl=False)
        click.secho(f"{session.session_id}", fg="green")
        click.secho(f"Session URL: ", nl=False)
        click.secho(f"{session.session_url}", fg="green")

    if web:
        click.launch(session.session_url)
        return
    if use_console:
        c = console.Console(client, session=session)
        c.run(message)
    else:
        if message:
            response = session.query(message)
            print(response)
        if web:
            if keep:
                click.secho(
                    "Warning: --keep is ignored when --web is specified.",
                    fg="yellow",
                )
            click.launch(session.session_url)
            return
        elif not message:
            raise click.BadParameter(
                "You must specify either --console, --web, or a query message"
            )

    if not keep:
        if verbose:
            click.echo(f"Deleting session ", nl=False)
            click.secho(f"{session.session_id}", fg="green")
        session.delete_session()


@session.command("new", help="Creates a new session and opens it.")
@click.argument("message", required=False)
@web_option
@click.option(
    "-a",
    "--agent",
    required=False,
    callback=validate_agent_exists,
    help="A specific agent to talk to. If unset, `fixie` is used.",
)
@click.option("-v", "--verbose", is_flag=True, help="Enable verbose output.")
@click.pass_context
def new_session(ctx, agent, web, message, verbose):
    run_session(ctx.obj.client, agent=agent, message=message, web=web, use_console=True)


@session.command("open", help="Opens a session.")
@click.argument("session_id", required=True)
@click.argument("message", required=False)
@web_option
@click.option(
    "-a",
    "--agent",
    required=False,
    callback=validate_agent_exists,
    help="A specific agent to talk to. If unset, `fixie` is used.",
)
@click.option("-v", "--verbose", is_flag=True, help="Enable verbose output.")
@click.pass_context
def open_session(ctx, session_id: str, web, agent, message, verbose):
    run_session(
        ctx.obj.client,
        session_id=session_id,
        agent=agent,
        message=message,
        web=web,
        use_console=True,
    )


@session.command("query", help="Runs a single query and exit.")
@click.argument("message", required=True)
@web_option
@click.option("--keep", is_flag=True, help="Do not delete the session after use.")
@click.option("-v", "--verbose", is_flag=True, help="Enable verbose output.")
@click.option(
    "-a",
    "--agent",
    required=False,
    callback=validate_agent_exists,
    help="A specific agent to talk to. If unset, `fixie` is used.",
)
@click.pass_context
def query(ctx, agent, web, keep, verbose, message):
    run_session(
        ctx.obj.client,
        agent=agent,
        message=message,
        web=web,
        keep=keep,
        verbose=verbose,
        use_console=False,
    )



from typing import Optional

from pydantic import dataclasses as pydantic_dataclasses


@pydantic_dataclasses.dataclass
class LlmSettings:
    """LLM settings for agents.

    These settings influence how agent responses are generated by the underlying LLM.

    Args:
        model Optional[str]: The name of the LLM to use.
            See https://docs.fixie.ai/agents/#agent-default-model-and-model-params for supported
            values.
        temperature Optional[float]: The sampling temperature to use, between 0 and 2.
            Higher values like 0.8 will make the output more random, while lower values like
            0.2 will make it more focused and deterministic. The precise behavior is
            model-specific.
        maximum_tokens Optional[int]: The maximum number of tokens to generate in responses.
            One token is typically a subword. The exact token definition depends on the model.
    """

    model: Optional[str] = None
    temperature: Optional[float] = None
    maximum_tokens: Optional[int] = None



import io
import os.path
import re
from typing import Any, Dict, List, Optional

import prompt_toolkit
import prompt_toolkit.history
import requests
from PIL import Image
from rich import console as rich_console
from rich.markdown import Markdown

from fixieai import FixieClient
from fixieai.client.client import Session

HISTORY_FILE = "~/.config/fixie/history.log"
textconsole = rich_console.Console(soft_wrap=True)
PROMPT = "fixie 🦊❯ "


class Console:
    """A simple console interface for Fixie."""

    def __init__(
        self,
        client: FixieClient,
        session: Session,
        history_file: str = HISTORY_FILE,
    ):
        self._client = client
        self._session = session
        history_file = os.path.expanduser(history_file)
        os.makedirs(os.path.dirname(history_file), exist_ok=True)
        self._history_file = history_file
        self._response_index = 0

    def run(
        self,
        initial_message: Optional[str] = None,
    ) -> None:
        """Run the console application."""

        textconsole.print("[blue]Welcome to Fixie!")
        textconsole.print(f"Connected to: {self._session.session_url}")

        # Show what's already in the session thus far.
        for message in self._session.get_messages_since_last_time():
            self._show_message(message, show_user_message=True)

        history = prompt_toolkit.history.FileHistory(self._history_file)
        if initial_message:
            prompt_toolkit.print_formatted_text(f"{PROMPT}{initial_message}")
            history.append_string(initial_message)
            self._query(initial_message)

        while True:
            in_text = prompt_toolkit.prompt(
                PROMPT,
                history=history,
                auto_suggest=prompt_toolkit.auto_suggest.AutoSuggestFromHistory(),
            )
            self._query(in_text)

    def _query(self, in_text: str) -> None:
        with textconsole.status("Working...", spinner="bouncingBall"):
            try:
                for message in self._session.run(in_text):
                    self._show_message(message)
            except requests.exceptions.HTTPError as e:
                textconsole.print(f"🚨 {e}")
                return

    def _show_message(self, message: Dict[str, Any], show_user_message: bool = False):
        """Shows a message dict from FixieClient.

        If show_user_message is set, the user messages are also printed with the PROMPT.
        This option is useful for showing previous messages in the chat when connecting
        to a session.
        """
        sender_handle = (
            message["sentBy"]["handle"] if message["sentBy"] else "<unknown>"
        )
        message_text = message["text"]

        if message["type"] == "query" and sender_handle == "user":
            if show_user_message:
                textconsole.print(Markdown(PROMPT + message_text))
        elif message["type"] != "response":
            textconsole.print(
                Markdown(f"   @{sender_handle}: " + message_text, style="dim")
            )
        else:
            self._response_index += 1
            textconsole.print(Markdown(f"{self._response_index}❯ " + message_text))
            self._show_embeds(message["text"])

    def _show_embeds(self, message: str):
        """Shows embeds referenced in `message_text`."""
        # Check embed references in message (denoted by #id).
        embed_ids = _extract_embed_refs(message)
        if not embed_ids:
            return
        # Get a dict of all embed_id -> embeds in the session.
        embeds = {
            embed_dict["key"]: embed_dict["embed"]
            for embed_dict in self._session.get_embeds()
        }
        # Show what we can find.
        for embed_id in embed_ids:
            if embed_id not in embeds:
                textconsole.print(
                    f"   [dim]embed #{embed_id} not found in session[/]", style="red"
                )
                continue
            _show_embed(
                embeds[embed_id]["url"],
                embeds[embed_id]["contentType"],
            )


def _extract_embed_refs(message: str) -> List[int]:
    """Returns a list of embed ids referenced in `message_text`."""
    embed_refs = []
    odd_number_of_sharps = r"(?<!#)(##)*#(?!#)"
    for match in re.finditer(odd_number_of_sharps + r"(?P<embed_num>\d+)", message):
        embed_refs.append(int(match.group("embed_num")))
    return embed_refs


def _show_embed(url: str, content_type: str):
    if content_type.startswith("image/"):
        response = requests.get(url)
        response.raise_for_status()
        image = Image.open(io.BytesIO(response.content))
        image.show()



import doctest
import re

import pytest

from fixieai import constants
from fixieai.agents import exceptions
from fixieai.agents import token
from fixieai.agents import user_storage

FAKE_AGENT_ID = "fake-agent"
FAKE_ACCESS_TOKEN = "fake-access-token"


class MockUserStorageService:
    """Mocks /api/userstorage behavior."""

    def __init__(self):
        self._data = {}

    def _parse_request(self, request):
        assert request.headers["Authorization"] == f"Bearer {FAKE_ACCESS_TOKEN}"
        prefix = f"/api/userstorage/{FAKE_AGENT_ID}"
        assert request.path.startswith(prefix)
        key = request.path[len(prefix) + 1 :]
        return key

    def post(self, request, context):
        key = self._parse_request(request)
        assert key
        self._data[key] = request.json()["data"]
        return {"msg": "success"}

    def get(self, request, context):
        key = self._parse_request(request)
        if key:
            return {"data": self._data[key]}
        else:
            return [{"key": key, "value": value} for key, value in self._data.items()]

    def head(self, request, context):
        key = self._parse_request(request)
        assert key
        try:
            self._data[key]
        except KeyError:
            context.status_code = 400
            return {"msg": "key doesn't exist"}
        else:
            context.status_code = 200
            return {"msg": "key exists"}

    def delete(self, request, context):
        key = self._parse_request(request)
        assert key
        del self._data[key]
        return {"msg": "success"}


@pytest.fixture
def mock_user_storage_urls(requests_mock):
    mock_storage_service = MockUserStorageService()
    user_storage_path_re = re.compile(
        "^" + re.escape(constants.FIXIE_USER_STORAGE_URL) + "/"
    )
    requests_mock.post(user_storage_path_re, json=mock_storage_service.post)
    requests_mock.get(user_storage_path_re, json=mock_storage_service.get)
    requests_mock.head(user_storage_path_re, json=mock_storage_service.head)
    requests_mock.delete(user_storage_path_re, json=mock_storage_service.delete)


VALUES_TO_TEST = [
    1,
    2.4,
    True,
    b"binary",
    "simple_str",
    {"key": "value"},
    {"key1": {"key2": ["value1", True, b"binary2", None, 4.3]}},
]


@pytest.mark.parametrize("test_value", VALUES_TO_TEST)
def test_user_storage(mock_user_storage_urls, test_value):
    storage = user_storage.UserStorage(
        token.VerifiedTokenClaims(
            FAKE_AGENT_ID, is_anonymous=False, token=FAKE_ACCESS_TOKEN
        )
    )

    # Add the value and test its existence
    storage["key"] = test_value
    assert "key" in storage
    assert len(storage) == 1
    assert storage["key"] == test_value
    assert list(storage) == ["key"]

    del storage["key"]
    assert "key" not in storage
    assert len(storage) == 0
    with pytest.raises(KeyError):
        _ = storage["key"]

    val = storage.get("key", "default value")
    assert val == "default value"


def test_doctest(mock_user_storage_urls):
    doctest.testmod(user_storage, raise_on_error=True)


def test_userstorage_fails_for_anonymous_users():
    with pytest.raises(exceptions.AgentException):
        user_storage.UserStorage(
            token.VerifiedTokenClaims(
                FAKE_AGENT_ID, is_anonymous=True, token=FAKE_ACCESS_TOKEN
            )
        )



from typing import Optional

import fastapi
from fastapi import testclient
from starlette import responses

from fixieai.agents import openai_proxy


def test_authorization_token_middleware():
    FAKE_TOKEN = "fixie-test-token"
    app = fastapi.FastAPI()

    @app.get("/")
    async def test_endpoint(expected_token: Optional[str] = None):
        assert openai_proxy._current_request_token.get() == expected_token

    @app.get("/stream")
    async def test_stream(count: int, expected_token: Optional[str] = None):
        def _gen():
            for _ in range(count):
                assert openai_proxy._current_request_token.get() == expected_token
                yield b"."

        return responses.StreamingResponse(_gen(), status_code=200)

    app.add_middleware(openai_proxy.OpenAIProxyRequestTokenForwarderMiddleware)

    client = testclient.TestClient(app)
    response = client.get(
        f"/?expected_token=test1",
        headers={"Authorization": f"Bearer test1"},
    )
    assert response.status_code == 200

    response = client.get(f"/")
    assert response.status_code == 200

    response = client.get(
        f"/?expected_token=test2",
        headers={"Authorization": f"Bearer test2"},
    )
    assert response.status_code == 200

    response = client.get("/stream?count=10")
    assert response.status_code == 200
    assert response.content == b"." * 10

    response = client.get(
        "/stream?count=10&expected_token=test-streaming",
        headers={"Authorization": f"Bearer test-streaming"},
    )
    assert response.status_code == 200
    assert response.content == b"." * 10



from __future__ import annotations

import contextlib
import dataclasses
import traceback

import fastapi
from fastapi.responses import JSONResponse

from fixieai.agents import api

UNHANDLED_EXCEPTION_RESPONSE_TEXT = (
    "I'm sorry, an error occurred while processing your request."
)


class AgentException(Exception):
    """An exception raised by an Agent."""

    def __init__(
        self,
        response_message: str,
        error_code: str,
        error_message: str,
        http_status_code: int = 500,
        **details,
    ):
        super().__init__(error_message)
        self._response_message = response_message
        self._http_status_code = http_status_code
        self._error_code = error_code
        self._error_message = error_message
        self._details = details

    def to_response(self) -> api.AgentResponse:
        """Converts the AgentException into an AgentResponse."""

        # Add exception details to the error details.
        error = api.AgentError(
            code=self._error_code,
            message=self._error_message,
            details={
                "traceback": traceback.format_exception(
                    type(self), self, self.__traceback__
                ),
                **self._details,
            },
        )

        return api.AgentResponse(
            api.Message(self._response_message),
            error=error,
        )

    @classmethod
    def fastapi_handler(cls, request: fastapi.Request, exc: AgentException):
        """A FastAPI exception handler for AgentExceptions."""
        return JSONResponse(
            dataclasses.asdict(exc.to_response()),
            status_code=exc._http_status_code,
        )

    @staticmethod
    @contextlib.contextmanager
    def exception_remapper():
        """Remaps unhandled exceptions into AgentExceptions."""
        try:
            yield
        except AgentException:
            raise
        except Exception as exc:
            raise AgentException(
                response_message=UNHANDLED_EXCEPTION_RESPONSE_TEXT,
                error_code="ERR_AGENT_UNHANDLED_EXCEPTION",
                error_message="An unhandled exception occurred while processing the request.",
                http_status_code=500,
            ) from exc



import click
from gql.transport import exceptions as gql_exceptions

from fixieai.cli.auth import oauth_flow
from fixieai.cli.auth import user_config
from fixieai.client import client


def validate_not_authed(ctx, param, force):
    """Click option callback for "fixie auth --force" that validates user is not
    authenticated if "--force" is not set."""
    if force:
        # --force is set
        return
    try:
        auth_token = user_config.load_config().auth_token
        if auth_token is not None:
            username = _get_username(auth_token)
            click.echo("Already authenticated as ", nl=False)
            click.secho(username, fg="green", nl=False, bold=True)
            click.echo(". Set --force to force re-authentication.")
            ctx.exit()
    except (FileNotFoundError, gql_exceptions.TransportQueryError):
        pass


def _get_username(auth_token: str) -> str:
    """Gets the username for an auth token."""
    return client.FixieClient(auth_token).get_current_username()


@click.command("auth", help="Authorizes `fixie` to access Fixie platform.")
@click.option(
    "--force",
    is_flag=True,
    help="Forces authentication, even if the user is authenticated.",
    callback=validate_not_authed,
    expose_value=False,
)
def auth():
    fixie_auth_token = oauth_flow.oauth_flow()
    try:
        config = user_config.load_config()
    except FileNotFoundError:
        config = user_config.UserConfig()

    config.auth_token = fixie_auth_token
    user_config.save_config(config)

    username = _get_username(fixie_auth_token)
    click.secho("Success! You're authenticated as ", fg="green", bold=True, nl=False)
    click.secho(username, fg="magenta", nl=False, bold=True)
    click.secho(".")


@click.command("user", help="Displays information on the authenticated user.")
@click.pass_context
def user(ctx):
    current_user = ctx.obj.client.get_current_user()
    click.secho("You are authenticated to ", nl=False)
    click.secho(ctx.obj.client.url, fg="yellow", nl=False)
    click.secho(" as:")
    for key, value in current_user.items():
        click.secho(key, fg="green", nl=False)
        click.echo(f": {value}")



import dataclasses
import os
from typing import List, Optional

import dataclasses_json

from fixieai import constants
from fixieai.cli import utils


@dataclasses.dataclass
class EnvironmentSpecificAuthToken(utils.DataClassYamlMixin):
    """Represents an auth token bound to a specific environment."""

    url: str
    """The URL of the environment."""

    auth_token: str
    """The auth token for the environment."""


@dataclasses.dataclass
class UserConfig(utils.DataClassYamlMixin):
    """Represents user config at ~/.config/fixie/config.yaml"""

    default_auth_token: Optional[str] = dataclasses.field(
        default=None, metadata=dataclasses_json.config(field_name="fixie_api_key")
    )
    """The default auth token, will be used if no URL-specific auth token matches."""

    environment_auth_tokens: List[EnvironmentSpecificAuthToken] = dataclasses.field(
        default_factory=lambda: [],
    )
    """Environment-specific auth tokens."""

    @property
    def auth_token(self) -> Optional[str]:
        for environment_key in self.environment_auth_tokens:
            if environment_key.url == constants.FIXIE_API_URL:
                return environment_key.auth_token

        # Otherwise return the default auth token
        return self.default_auth_token

    @auth_token.setter
    def auth_token(self, value: str):
        if self.auth_token != value:
            # Append it to the environment-specific tokens and make it the default
            self.environment_auth_tokens.append(
                EnvironmentSpecificAuthToken(constants.FIXIE_API_URL, value)
            )
            self.default_auth_token = value


def load_config(path: Optional[str] = None) -> UserConfig:
    """Loads UserConfig from the given path, or its default path.

    Args:
        path: Optional path to load config from. By default, config is loaded from
            "~/.config/fixie/config.yaml".
    """
    if path is None:
        path = os.path.expanduser("~/.config/fixie/config.yaml")
    with open(path, "r") as fp:
        return UserConfig.from_yaml(fp)


def save_config(user_config: UserConfig, path: Optional[str] = None):
    """Saves the given AgentConfig to the given path, or its default path.

    Args:
        user_config: The UserConfig object to save into file.
        path: Optional path to save. By default, it's saved to the default path at
            "~/.config/fixie/config.yaml".
    """
    if path is None:
        path = os.path.expanduser("~/.config/fixie/config.yaml")
        os.makedirs(os.path.expanduser("~/.config/fixie"), exist_ok=True)
    with open(path, "w") as fp:
        fp.write(user_config.to_yaml())



from __future__ import annotations

import enum
import re
from typing import TYPE_CHECKING, Callable, Optional, Set

if TYPE_CHECKING:
    from fixieai.agents import code_shot

ODD_NUM_POUNDS = re.compile(r"(?<!#)(##)*#(?!#)")
EMBED_REF = re.compile(ODD_NUM_POUNDS.pattern + r"(?P<embed_key>\w+)(?!\w)")


def strip_prompt_lines(agent: code_shot.CodeShotAgent):
    """Strips all prompt lines."""
    agent.base_prompt = _strip_all_lines(agent.base_prompt)
    for i, fewshot in enumerate(agent.few_shots):
        agent.few_shots[i] = _strip_all_lines(fewshot)


def validate_code_shot_agent(agent: code_shot.CodeShotAgent):
    """A client-side validation of few_shots and agent."""
    _validate_base_prompt(agent.base_prompt)
    for fewshot in agent.few_shots:
        _validate_few_shot_prompt(
            fewshot, agent.conversational, agent.is_valid_func_name
        )


def _strip_all_lines(prompt: str) -> str:
    prompt = prompt.strip()
    return "\n".join(line.strip() for line in prompt.splitlines())


def _validate_base_prompt(base_prompt: str):
    if base_prompt.endswith("\n") or base_prompt.startswith("\n"):
        raise ValueError(
            "base_prompt should not start or end in newlines. "
            f"base_prompt={base_prompt!r}."
        )
    whitespaces = (" ", "\t", "\r")
    prompt_lines = base_prompt.split("\n")
    bad_lines = [
        line
        for line in prompt_lines
        if line.startswith(whitespaces) or line.endswith(whitespaces)
    ]
    if bad_lines:
        raise ValueError(
            f"Some lines in the base prompt start or end in whitespaces: {bad_lines!r}."
        )


class FewshotLinePattern(enum.Enum):
    QUERY = re.compile(r"^Q:")
    AGENT_SAYS = re.compile(r"^Agent\[(?P<agent_id>\w+)] says:")
    FUNC_SAYS = re.compile(r"^Func\[(?P<func_name>\w+)] says:")
    ASK_AGENT = re.compile(r"^Ask Agent\[(?P<agent_id>\w+)]:")
    ASK_FUNC = re.compile(r"^Ask Func\[(?P<func_name>\w+)]:")
    RESPONSE = re.compile(r"^A:")
    NO_PATTERN: None = None

    @classmethod
    def match(cls, line: str) -> Optional[re.Match[str]]:
        """Returns a match from a FewshotLinePattern for a given line, or None if nothing matched."""
        if "\n" in line:
            raise ValueError(
                "Cannot get the pattern for a multi-line text. Patterns must be "
                "extracted one line at a time."
            )
        matches = [
            match
            for prompt_pattern in cls
            if prompt_pattern is not cls.NO_PATTERN
            and (match := prompt_pattern.value.match(line))
        ]
        if len(matches) > 1:
            raise RuntimeError(
                f"More than one pattern ({list(FewshotLinePattern(match.re) for match in matches)}) matched the line {line!r}."
            )
        elif len(matches) == 0:
            return None

        match = matches[0]
        if match.re is cls.QUERY.value:
            if match.end() == len(line):
                raise ValueError("A 'Q:' line cannot end without a query.")

        assert isinstance(match, re.Match)
        return match


def _validate_few_shot_prompt(
    prompt: str, conversational: bool, is_valid_func_name: Callable[[str], bool]
):
    """Validates 'prompt' as a correctly formatted few shot prompt."""
    lines = prompt.splitlines(False)

    # Check that no line starts or ends in a whitespace.
    whitespaces = (" ", "\t", "\r")
    bad_lines = [
        line
        for line in lines
        if line.startswith(whitespaces) or line.endswith(whitespaces)
    ]
    _assert(
        not bad_lines,
        f"Some lines in the fewshot start or end in whitespaces: {bad_lines!r}.",
        prompt,
    )

    # Check that it doesn't end with newline
    _assert(not prompt.endswith("\n"), "Fewshot ends with newline.", prompt)

    # Check that fewshot starts with a Q:, ends in an A:, and a Func says and Agent
    # says follows every Ask Func and Ask Agent.
    pattern_matches = [FewshotLinePattern.match(line) for line in lines]
    _assert(
        pattern_matches[0] is not None
        and pattern_matches[0].re is FewshotLinePattern.QUERY.value,
        "Fewshot must start with a 'Q:'",
        prompt,
    )
    last_pattern = FewshotLinePattern.QUERY
    known_embeds: Set[str] = set()

    for i, (match, line) in enumerate(zip(pattern_matches, lines)):
        pattern = (
            FewshotLinePattern(match.re)
            if match is not None
            else FewshotLinePattern.NO_PATTERN
        )

        if pattern is FewshotLinePattern.ASK_AGENT:
            assert match is not None
            next_match = (
                pattern_matches[i + 1] if i + 1 < len(pattern_matches) else None
            )
            _assert(
                next_match is not None
                and next_match.re is FewshotLinePattern.AGENT_SAYS.value
                and match.group("agent_id") == next_match.group("agent_id"),
                "Each 'Ask Agent' line must be immediately followed by an 'Agent says' line that references the same ",
                prompt,
            )

        if pattern is FewshotLinePattern.ASK_FUNC:
            assert match is not None
            next_match = (
                pattern_matches[i + 1] if i + 1 < len(pattern_matches) else None
            )
            _assert(
                next_match is not None
                and next_match.re is FewshotLinePattern.FUNC_SAYS.value
                and match.group("func_name") == next_match.group("func_name"),
                "Each 'Ask Func' line must be immediately followed by an 'Func says' line that references the same func.",
                prompt,
            )
            _assert(
                is_valid_func_name(match.group("func_name")),
                "Each 'Ask Func' line must reference a valid func name.",
                prompt,
            )

        if pattern is FewshotLinePattern.FUNC_SAYS:
            assert match is not None
            _assert(
                is_valid_func_name(match.group("func_name")),
                "Each 'Func says' line must reference a valid func name.",
                prompt,
            )

        if pattern is not FewshotLinePattern.NO_PATTERN:
            last_pattern = pattern

        # Check that any embeds are valid.
        for embed_match in EMBED_REF.finditer(line):
            key = embed_match.group("embed_key")
            if key not in known_embeds:
                _assert(
                    last_pattern
                    not in (
                        FewshotLinePattern.ASK_AGENT,
                        FewshotLinePattern.ASK_FUNC,
                        FewshotLinePattern.RESPONSE,
                    ),
                    "New embeds may not be introduced in Ask Agent, Ask Func, or A: lines.",
                    prompt,
                )
                known_embeds.add(key)

    _assert(
        last_pattern is FewshotLinePattern.RESPONSE,
        f"Fewshot must end with a 'A:' pattern, but it ends with {last_pattern}",
        prompt,
    )

    # Check that there's Q: and A: lines are interleaved.
    all_qa_lines = [
        "Q"
        if match.re is FewshotLinePattern.QUERY.value
        else "A"
        if match.re is FewshotLinePattern.RESPONSE.value
        else "x"
        for match in pattern_matches
        if match is not None
    ]
    qa_str = "".join(all_qa_lines)

    if conversational:
        _assert(
            re.match("^(Qx*A)+$", qa_str) is not None,
            "Each fewshot must have interleaved Q: and A: patterns when conversational=True",
            prompt,
        )
    else:
        _assert(
            re.match("^Qx*A$", qa_str) is not None,
            "Each fewshot must have exactly one Q: and A: pattern when conversational=False",
            prompt,
        )


def _assert(condition: bool, msg: str, prompt: str):
    if not condition:
        raise ValueError(f"{msg} in few-shot prompt: {prompt!r}.")



import contextlib
import functools
from typing import BinaryIO, Optional

from requests_toolbelt.multipart.encoder import MultipartEncoder  # type: ignore


@contextlib.contextmanager
def patched_gql_file_uploader(
    file_like: Optional[BinaryIO], name: str, content_type: str
):
    """Applies a temporary patch to allow setting name and Content-Type on uploads via `gql`.

    Future versions of `gql` will support setting Content-Type, but even with that support
    `gql` requires that the stream being uploaded is named and FixieClient does not. So this
    patches the underlying multipart encoder to assign the name and Content-Type to the specified
    stream.
    """

    if file_like is None:
        yield
        return

    @functools.wraps(MultipartEncoder.__init__)
    def _patched_init(self, fields, *args, **kwargs):
        patched_fields = {
            k: (
                (
                    name,
                    file_like,
                    content_type,
                )
                if len(v) == 2 and v[1] == file_like
                else v
            )
            for k, v in fields.items()
        }
        _patched_init.__wrapped__(self, patched_fields, *args, **kwargs)  # type: ignore

    try:
        MultipartEncoder.__init__ = _patched_init
        yield
    finally:
        MultipartEncoder.__init__ = _patched_init.__wrapped__  # type: ignore



from __future__ import annotations

import abc
import dataclasses
import functools
import json
import os
import re
import warnings
from typing import Callable, Dict, Optional, TypeVar

import fastapi
import fastapi.security
import jwt
import starlette.responses
import uvicorn
import yaml

from fixieai import constants
from fixieai.agents import agent_func
from fixieai.agents import api
from fixieai.agents import corpora
from fixieai.agents import exceptions
from fixieai.agents import metadata as agent_metadata
from fixieai.agents import oauth
from fixieai.agents import token

# Regex that controls what Func names are allowed.
ACCEPTED_FUNC_NAMES = re.compile(r"^\w+$")

_RESPONSE_CHUNK_SIZE = 1024


class AgentBase(abc.ABC):
    """Base class for Fixie agents."""

    def __init__(
        self,
        oauth_params: Optional[oauth.OAuthParams] = None,
    ):
        self.oauth_params = oauth_params
        self._funcs: Dict[str, agent_func.AgentQueryFunc] = {}
        self._corpus_funcs: Dict[str, agent_func.AgentCorpusFunc] = {}
        self._jwks_client = jwt.PyJWKClient(constants.FIXIE_JWKS_URL)
        self._allowed_agent_id = os.getenv("FIXIE_ALLOWED_AGENT_ID")
        if self._allowed_agent_id is None:
            warnings.warn(
                "No allowed agent ID was specified, so your agent will accept requests intended for any agent. "
                "Ensure that the FIXIE_ALLOWED_AGENT_ID variable is set to correct this."
            )

        if oauth_params is not None:
            # Register default Funcs.
            self.register_func(_oauth)

    @abc.abstractmethod
    def metadata(self) -> agent_metadata.Metadata:
        """Returns metadata about how the agent should be interacted with."""

    @abc.abstractmethod
    def validate(self):
        """Performs any validation after all funcs are registered."""

    def serve(self, host: str = "0.0.0.0", port: int = 8181):
        """Starts serving the current agent at `{host}:{port}` via uvicorn.

        Args:
            host: The address to start listening at.
            port: The port number to start listening at.
        """
        uvicorn.run(self.app(), host=host, port=port)

    def app(self) -> fastapi.FastAPI:
        """Returns a fastapi.FastAPI application that serves the agent."""
        fast_api = fastapi.FastAPI()
        fast_api.include_router(self.api_router())
        fast_api.add_exception_handler(
            exceptions.AgentException, exceptions.AgentException.fastapi_handler
        )

        return fast_api

    def api_router(self) -> fastapi.APIRouter:
        """Returns a fastapi.APIRouter object that serves the agent."""
        router = fastapi.APIRouter()
        router.add_api_route("/", self._handshake, methods=["GET"])
        router.add_api_route("/{func_name}", self._serve_func, methods=["POST"])
        router.add_api_route(
            "/corpus/{corpus_func_name}", self._serve_corpus_func, methods=["POST"]
        )
        return router

    def is_valid_func_name(self, func_name: str) -> bool:
        """Indicates if the given func name is valid (either registered or built-in).

        Args:
            func_name: The func name to check
        """
        return func_name in self._funcs or func_name.startswith("fixie_")

    def register_func(
        self, func: Optional[Callable] = None, *, func_name: Optional[str] = None
    ) -> Callable:
        """A function decorator to register `Func`s with this agent.

        This decorator will not change the callable itself.

        Usage:

            agent = CodeShotAgent(base_prompt, few_shots)

            @agent.register_func
            def func(query):
                ...

        Optional Decorator Args:
            func_name: Optional function name to register this function by. If unset,
                the function name will be used.
        """
        if func is None:
            # Func is not passed in. It's the decorator being created.
            return functools.partial(self.register_func, func_name=func_name)
        non_null_func: Callable = func

        AgentBase._register_func_internal(
            func,
            func_name,
            self._funcs,
            lambda: agent_func.AgentQueryFunc.create(
                non_null_func,
                self.oauth_params,
                default_message_type=api.Message,
                allow_generator=False,
            ),
        )
        return func

    def register_corpus_func(
        self,
        func: Optional[
            Callable[[corpora.CorpusRequest], corpora.CorpusResponse]
        ] = None,
        *,
        func_name: Optional[str] = None,
    ) -> Callable:
        """A function decorator to register `Func`s used for loading a custom
        corpus with this agent.

        This decorator will not change the callable itself.

        Usage:

            agent = CodeShotAgent(base_prompt, few_shots)

            @agent.register_corpus_func
            def func(request: CorpusRequest) -> CorpusResponse:
                ...

        Optional Decorator Args:
            func_name: Optional function name to register this function by. If unset,
                the function name will be used.
        """
        if func is None:
            # Func is not passed in. It's the decorator being created.
            return functools.partial(self.register_corpus_func, func_name=func_name)
        non_null_func: Callable[[corpora.CorpusRequest], corpora.CorpusResponse] = func

        AgentBase._register_func_internal(
            func,
            func_name,
            self._corpus_funcs,
            lambda: agent_func.AgentCorpusFunc.create(non_null_func),
        )
        return func

    T = TypeVar("T")

    @staticmethod
    def _register_func_internal(
        func: Callable,
        func_name: Optional[str],
        registry: Dict[str, T],
        create: Callable[[], T],
    ):
        if func_name is not None:
            if not ACCEPTED_FUNC_NAMES.fullmatch(func_name):
                raise ValueError(
                    f"Function names may only be alphanumerics, got {func_name!r}."
                )
        else:
            func_name = func.__name__

        if func_name in registry:
            raise ValueError(f"Func[{func_name}] is already registered with agent.")

        registry[func_name] = create()

    def _handshake(
        self,
        credentials: fastapi.security.HTTPAuthorizationCredentials = fastapi.Depends(
            fastapi.security.HTTPBearer()
        ),
    ) -> fastapi.Response:
        """Returns the agent's metadata in YAML format."""
        token_claims = token.VerifiedTokenClaims.from_token(
            credentials.credentials, self._jwks_client, self._allowed_agent_id
        )
        if token_claims is None:
            raise fastapi.HTTPException(status_code=403, detail="Invalid token")

        metadata = self.metadata()
        yaml_content = yaml.dump(dataclasses.asdict(metadata))

        # Use a streaming response because metadata can be very large.
        chunks = (
            yaml_content[start : start + _RESPONSE_CHUNK_SIZE]
            for start in range(0, len(yaml_content), _RESPONSE_CHUNK_SIZE)
        )
        return starlette.responses.StreamingResponse(
            chunks,
            media_type="application/yaml",
        )

    def _serve_func(
        self,
        func_name: str,
        query: api.AgentQuery,
        credentials: fastapi.security.HTTPAuthorizationCredentials = fastapi.Depends(
            fastapi.security.HTTPBearer()
        ),
    ) -> api.AgentResponse:
        """Verifies the request is a valid request from Fixie, and dispatches it to
        the appropriate function.
        """
        token_claims = self._check_credentials(credentials)

        try:
            pyfunc = self._funcs[func_name]
        except KeyError:
            raise exceptions.AgentException(
                response_message=f"I'm sorry, Func[{func_name}] is not defined.",
                error_code="ERR_FUNC_NOT_DEFINED",
                error_message="The function was not defined.",
                http_status_code=404,
                func_name=func_name,
                available_funcs=list(self._funcs.keys()),
            )

        output = pyfunc(query, token_claims)

        # Streaming not allowed for funcs (yet).
        return next(iter(output))

    def _serve_corpus_func(
        self,
        corpus_func_name: str,
        request: corpora.CorpusRequest,
        credentials: fastapi.security.HTTPAuthorizationCredentials = fastapi.Depends(
            fastapi.security.HTTPBearer()
        ),
    ) -> fastapi.responses.JSONResponse:
        """Verifies the request is a valid request from Fixie, and dispatches it to
        the appropriate corpus function.
        """
        token_claims = self._check_credentials(credentials)

        try:
            pyfunc = self._corpus_funcs[corpus_func_name]
        except KeyError:
            raise exceptions.AgentException(
                response_message=f"I'm sorry, corpus func[{corpus_func_name}] is not defined.",
                error_code="ERR_FUNC_NOT_DEFINED",
                error_message="The function was not defined.",
                http_status_code=404,
                func_name=corpus_func_name,
                available_funcs=list(self._corpus_funcs.keys()),
            )

        output = pyfunc(request, token_claims)

        result = next(iter(output))
        # Use dataclasses_json instead of FastAPI's default json encoder. This
        # allows us to use our own bytes encoder instead of the default, which
        # always uses utf-8 and therefore wouldn't allow some characters.
        return fastapi.responses.JSONResponse(result.to_dict())

    def _check_credentials(
        self, credentials: fastapi.security.HTTPAuthorizationCredentials
    ) -> token.VerifiedTokenClaims:
        token_claims = token.VerifiedTokenClaims.from_token(
            credentials.credentials, self._jwks_client, self._allowed_agent_id
        )
        if token_claims is None:
            raise fastapi.HTTPException(status_code=403, detail="Invalid token")
        return token_claims


def _oauth(query: api.Message, oauth_handler: oauth.OAuthHandler) -> str:
    """Serves Func[_oauth] which is used upon auth redirect callback."""
    auth_request = json.loads(query.text)
    state = auth_request["state"]
    code = auth_request["code"]
    oauth_handler.authorize(state, code)
    return "Authorization successful!"



import base64
import dataclasses
from typing import List, Optional

import dataclasses_json.core
from deprecated import deprecated
from pydantic import dataclasses as pydantic_dataclasses


@pydantic_dataclasses.dataclass
class CorpusRequest(dataclasses_json.DataClassJsonMixin):
    """A request for some piece of the agent's corpus.

    In addition to returning documents, each response may expand the corpus
    space in one or both of two dimensions: new partitions and next pages.

    Partitions are non-overlapping subsets of a corpus which may be loaded in
    parallel by Fixie. A response's new partitions will be ignored if
    previously included in another response.

    When a response includes a page of documents, that page may indicate that
    another page is available in the same partition. Pages are always loaded
    serially in order. The partition is completed when a response has a page
    with no next_page_token.

    Agents will always receive a first request with the default (unnamed)
    partition and no page_token. Subsequent requests depend on prior responses
    and will always include at least one of those fields.

    Examples:
        Simple handful of documents:

            When receiving the initial request, the agent responds with a page
            of documents. This could include a next_page_token for more
            documents in the single default partition if needed.

        Web crawl:

            Each URL corresponds to a partition and the agent never returns
            tokens. The initial response only includes partitions, one for each
            root URL to crawl. Each subsequent request includes the partition
            (the URL) and the corresponding response contains a page with a
            single document - the resource at that URL. If the document links
            to other resources that should be included in the corpus, the
            response also contains those URLs as new partitions. The process
            repeats for all partitions until there are no known incomplete
            partitions or until crawl limits are reached.

        Database:

            Consider a database with a parent table keyed by parent_id and an
            interleaved child table keyed by (parent_id, child_id) whose rows
            correspond to corpus documents. This agent will use tokens that
            encode a read timestamp (for consistency) and an offset to be used
            in combination with a static page size.

            Upon receiving the initial CorpusRequest, the agent chooses a
            commit timestamp to use for all reads and returns a partition for
            each parent_id along with a first_page_token indicating the chosen
            read timestamp and an offset of 0.

            For each partition, the agent then receives requests with the
            partition (a parent_id) and a page token (the read timestamp and
            offest). It responds with documents corresponding to the next page
            size child rows within the given parent. If more children exist,
            the response includes a next_page_token with the same read
            timestamp and an incremented offset. This repeats until there are
            no more children, at which point the response has no
            next_page_token and the partition is complete.

            Note: Including multiple parent_ids in each partition would also
                work and would be an effective way to limit parallelism if
                desired.

    Attributes:
        partition: The partition of the corpus that should be read. This will
            be empty for the initial request, indicating the default partition.
            For subsequent requests, it will either be the name of a partition
            returned by a previous request or empty if the default partition
            contains multiple pages for this agent.
        page_token: A token for paginating results within a corpus partition.
            If present, this will be echoed from a previous response.
    """

    partition: Optional[str] = None
    page_token: Optional[str] = None


@pydantic_dataclasses.dataclass
class CorpusPartition:
    """An identifier for a subset of a corpus, along with an optional
    token to use when loading its first page. Each partition will only be
    loaded once during a single crawl. If multiple responses include the same
    partition, the token of the first received response will be used."""

    partition: str
    first_page_token: Optional[str] = None


@pydantic_dataclasses.dataclass
class CorpusDocument:
    """Some meaningful item of data from a corpus. This could be an HTML page,
    a PDF, or a raw string of text (among others). Fixie will handle parsing
    and chunking this document so that appropriately sized chunks can be
    included in LLM requests.

    Note: If custom parsing is desired, agents are free to implement their own
    parsing to return documents with text/plain mime_types instead of whatever
    they fetch natively. Fixie will not alter text/plain documents prior to
    chunking."""

    source_name: str
    content: bytes = dataclasses.field(
        metadata=dataclasses_json.config(
            encoder=lambda c: base64.b64encode(c).decode(),
            decoder=lambda c: base64.b64decode(c.encode()),
        )
    )
    encoding: str = "UTF-8"
    mime_type: str = "text/plain"

    @property
    def text(self) -> str:
        return self.content.decode(self.encoding)


@pydantic_dataclasses.dataclass
class CorpusPage:
    """A page of CorpusDocuments. In addition to the documents themselves, a
    page may include a continuation token for fetching the next page (in the
    same partition). Omitting a token implies that this is the last page."""

    documents: List[CorpusDocument]
    next_page_token: Optional[str] = None


@pydantic_dataclasses.dataclass
class CorpusResponse(dataclasses_json.DataClassJsonMixin):
    """A response to a CorpusRequest. See CorpusRequest for details."""

    partitions: Optional[List[CorpusPartition]] = None
    page: Optional[CorpusPage] = None


@pydantic_dataclasses.dataclass
class CustomCorpus:
    """A custom corpus for a Fixie CodeShot Agent. This uses a registered
    corpus func to load documents from an arbitrary source."""

    func_name: str


@deprecated(reason="Use register_corpus_func for custom document loading.")
@pydantic_dataclasses.dataclass
class DocumentLoader:
    """Deprecated. This doesn't do anything."""

    name: str


@pydantic_dataclasses.dataclass
class UrlDocumentCorpus:
    """URL Document corpus for a Fixie CodeShot Agent."""

    urls: List[str]
    """URLs to load documents from. A trailing wildcard (e.g., https://example.com/*),
    can be used to load all documents from a site."""

    exclude_patterns: Optional[List[str]] = None
    """A list of wildcard patterns to exclude from crawled URLs (e.g., */no_crawl/*)."""

    auth_token_func: Optional[str] = None

    loader: Optional[DocumentLoader] = None  # Deprecated.


DocumentCorpus = UrlDocumentCorpus  # Deprecated: Prefer UrlDocumentCorpus



import dataclasses
import os
from typing import Optional

from fixieai.cli import utils


@dataclasses.dataclass
class AgentConfig(utils.DataClassYamlMixin):
    """Represents an agent.yaml config file."""

    handle: str
    name: Optional[str] = None
    description: str = ""
    more_info_url: str = ""
    entry_point: str = "main:agent"
    deployment_url: Optional[str] = None
    public: Optional[bool] = None


def normalize_path(path: Optional[str] = None) -> str:
    """Normalizes paths to an AgentConfig.

    Args:
        path: Optional path to either a directory or YAML file. If unspecified,
            will return "agent.yaml".
    """
    if not path:
        path = "agent.yaml"
    elif os.path.isdir(path):
        path = os.path.join(path, "agent.yaml")

    return path


def load_config(path: Optional[str] = None) -> AgentConfig:
    """Loads AgentConfig from the given path, or its default path.

    Args:
        path: Optional path to load config from. By default, config is loaded from
            "{cwd}/agent.yaml".
    """
    path = normalize_path(path)
    with open(path, "r") as fp:
        return AgentConfig.from_yaml(fp)


def save_config(agent_config: AgentConfig, path: Optional[str] = None):
    """Saves the given AgentConfig to the given path, or its default path.

    Args:
        agent_config: The AgentConfig object to save into file.
        path: Optional path to save. By default, it's saved to the default path at
            "{cwd}/agent.yaml".
    """
    path = normalize_path(path)
    with open(path, "w") as fp:
        fp.write(agent_config.to_yaml())



from __future__ import annotations

import datetime
import warnings
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from gql import gql

if TYPE_CHECKING:
    import fixieai.client as fixie_client

    FixieClient = fixie_client.FixieClient
else:
    FixieClient = Any


class Agent:
    """Provides an interface to the Fixie GraphQL Agent API.

    Args:
        client: The FixieClient instance to use.
        agent_id: The Agent ID, e.g., "fixie/calc", or handle, e.g., "dice".
    """

    def __init__(
        self,
        client: FixieClient,
        agent_id: str,
    ):
        self._client = client
        self._gqlclient = self._client.gqlclient

        self._agent_id = agent_id
        self._owner: Optional[str] = None
        if "/" in agent_id:
            self._owner, self._handle = agent_id.split("/")
        else:
            self._owner = self._client.get_current_username()
            self._handle = agent_id

        self._metadata: Optional[Dict[str, Any]] = None
        try:
            self._metadata = self.get_metadata()
        except:
            self._metadata = None

    @property
    def agent_id(self) -> str:
        """Return the agentId for this Agent."""
        return self._agent_id

    @property
    def handle(self) -> str:
        """Return the handle for this Agent."""
        return self._handle

    @property
    def valid(self) -> bool:
        """Return whether this Agent is valid."""
        return self._metadata is not None

    @property
    def name(self) -> Optional[str]:
        """Return the name for this Agent."""
        if self._metadata is None:
            return None
        name = self._metadata["name"]
        assert name is None or isinstance(name, str)
        return name

    @property
    def description(self) -> Optional[str]:
        """Return the description for this Agent."""
        if self._metadata is None:
            return None
        description = self._metadata["description"]
        assert description is None or isinstance(description, str)
        return description

    @property
    def queries(self) -> Optional[List[str]]:
        """Return the queries for this Agent."""
        if self._metadata is None:
            return None
        queries = self._metadata["queries"]
        assert queries is None or (
            isinstance(queries, list) and all(isinstance(q, str) for q in queries)
        )
        return queries

    @property
    def more_info_url(self) -> Optional[str]:
        """Return the more info URL for this Agent."""
        if self._metadata is None:
            return None
        more_info_url = self._metadata["moreInfoUrl"]
        assert more_info_url is None or isinstance(more_info_url, str)
        return more_info_url

    @property
    def published(self) -> Optional[bool]:
        """Return the published status for this Agent."""
        if self._metadata is None:
            return None
        published = self._metadata["published"]
        assert published is None or isinstance(published, bool)
        return published

    @property
    def owner(self) -> Optional[str]:
        """Return the owner of this Agent, which is either a username or an Organization handle."""
        if self._metadata is None:
            return None
        owner_username = self._metadata["owner"].get("username") or self._metadata[
            "owner"
        ].get("handle")
        assert owner_username is None or isinstance(owner_username, str)
        return owner_username

    @property
    def query_url(self) -> Optional[str]:
        """Return the query URL for this Agent."""
        if self._metadata is None:
            return None
        url = self._metadata["queryUrl"]
        assert url is None or isinstance(url, str)
        return url

    @property
    def func_url(self) -> Optional[str]:
        """Return the func URL for this Agent."""
        if self._metadata is None:
            return None
        url = self._metadata["funcUrl"]
        assert url is None or isinstance(url, str)
        return url

    @property
    def created(self) -> Optional[datetime.datetime]:
        """Return the creation timestamp for this Agent."""
        if self._metadata is None:
            return None
        ts = self._metadata["created"]
        if ts is not None:
            return datetime.datetime.fromisoformat(ts)
        else:
            return None

    @property
    def modified(self) -> Optional[datetime.datetime]:
        """Return the modification timestamp for this Agent."""
        if self._metadata is None:
            return None
        ts = self._metadata["modified"]
        if ts is not None:
            return datetime.datetime.fromisoformat(ts)
        else:
            return None

    def get_metadata(self) -> Dict[str, Any]:
        """Return metadata about this Agent."""
        query = gql(
            """
            query getAgentById($agentId: String!) {
                agentById(agentId: $agentId) {
                    agentId
                    handle
                    name
                    description
                    queries
                    moreInfoUrl
                    published
                    owner {
                        __typename
                        ... on UserType {
                            username
                        }
                        ... on OrganizationType {
                            handle
                        }
                    }
                    queryUrl
                    funcUrl
                    created
                    modified
                }
            }
        """
        )
        result = self._gqlclient.execute(
            query, variable_values={"agentId": f"{self._owner}/{self._handle}"}
        )
        if "agentById" not in result or result["agentById"] is None:
            raise ValueError(
                f"Cannot fetch agent metadata for {self._owner}/{self._handle}"
            )
        agent_dict = result["agentById"]

        assert isinstance(agent_dict, dict) and all(
            isinstance(k, str) for k in agent_dict.keys()
        )
        return agent_dict

    def create_agent(
        self,
        name: Optional[str],
        description: str,
        query_url: Optional[str] = None,
        func_url: Optional[str] = None,
        more_info_url: Optional[str] = None,
        published: Optional[bool] = None,
    ) -> str:
        """Create a new Agent with the given parameters."""
        query = gql(
            """
            mutation CreateAgent(
                $handle: String!,
                $name: String,
                $description: String!,
                $queryUrl: String,
                $funcUrl: String,
                $moreInfoUrl: String,
                $published: Boolean) {
                createAgent(
                    agentData: {
                        handle: $handle,
                        name: $name,
                        description: $description,
                        queryUrl: $queryUrl,
                        funcUrl: $funcUrl,
                        moreInfoUrl: $moreInfoUrl,
                        published: $published
                    }
                ) {
                    agent {
                        agentId
                    }
                }
            }
            """
        )

        variable_values: Dict[str, Any] = {"handle": self._handle}
        if name is not None:
            variable_values["name"] = name
        variable_values["description"] = description
        if query_url is not None:
            variable_values["queryUrl"] = query_url
        if func_url is not None:
            warnings.warn(
                "Setting func_url via create_agent is deprecated, use FixieClient.create_agent_revision instead."
            )
            variable_values["funcUrl"] = func_url
        if more_info_url is not None:
            variable_values["moreInfoUrl"] = more_info_url
        if published is not None:
            variable_values["published"] = published

        result = self._gqlclient.execute(query, variable_values=variable_values)
        if "createAgent" not in result or result["createAgent"] is None:
            raise ValueError(f"Failed to create Agent")
        assert isinstance(result["createAgent"], dict)
        assert isinstance(result["createAgent"]["agent"], dict)
        assert isinstance(result["createAgent"]["agent"]["agentId"], str)
        self._metadata = self.get_metadata()
        return result["createAgent"]["agent"]["agentId"]

    def update_agent(
        self,
        new_handle: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        query_url: Optional[str] = None,
        func_url: Optional[str] = None,
        more_info_url: Optional[str] = None,
        published: Optional[bool] = None,
    ) -> str:
        """Update the Agent with the given parameters."""
        query = gql(
            """
            mutation UpdateAgent(
                $handle: String!,
                $newHandle: String,
                $name: String,
                $description: String,
                $queryUrl: String,
                $funcUrl: String,
                $moreInfoUrl: String,
                $published: Boolean) {
                updateAgent(
                    agentData: {
                        handle: $handle,
                        newHandle: $newHandle,
                        name: $name,
                        description: $description,
                        queryUrl: $queryUrl,
                        funcUrl: $funcUrl,
                        moreInfoUrl: $moreInfoUrl,
                        published: $published
                    }
                ) {
                    agent {
                        agentId
                    }
                }
            }
            """
        )

        variable_values: Dict[str, Any] = {"handle": self._handle}
        if new_handle is not None:
            variable_values["newHandle"] = new_handle
        if name is not None:
            variable_values["name"] = name
        if description is not None:
            variable_values["description"] = description
        if query_url is not None:
            variable_values["queryUrl"] = query_url
        if func_url is not None:
            warnings.warn(
                "Setting func_url via update_agent is deprecated, use FixieClient.create_agent_revision instead."
            )
            variable_values["funcUrl"] = func_url
        if more_info_url is not None:
            variable_values["moreInfoUrl"] = more_info_url
        if published is not None:
            variable_values["published"] = published

        result = self._gqlclient.execute(query, variable_values=variable_values)
        if "updateAgent" not in result or result["updateAgent"] is None:
            raise ValueError(f"Failed to update Agent")
        assert isinstance(result["updateAgent"], dict)
        assert isinstance(result["updateAgent"]["agent"], dict)
        assert isinstance(result["updateAgent"]["agent"]["agentId"], str)
        self._metadata = self.get_metadata()
        if new_handle:
            self._handle = new_handle
        return result["updateAgent"]["agent"]["agentId"]

    def delete_agent(self) -> None:
        """Delete this Agent."""
        query = gql(
            """
            mutation DeleteAgent($owner: String!, $handle: String!) {
                deleteAgent(agentData: {owner: $owner, handle: $handle}) {
                    agent {
                        handle
                    }
                }
            }
        """
        )
        _ = self._gqlclient.execute(
            query, variable_values={"owner": self.owner, "handle": self._handle}
        )



#!/usr/bin/env python3

from typing import Optional

import click
import pkg_resources
import requests
from packaging import version

import fixieai
from fixieai import constants
from fixieai.cli.agent import commands as agent_commands
from fixieai.cli.auth import commands as auth_commands
from fixieai.cli.session import commands as session_commands


class CliContext:
    def __init__(self):
        self._client: Optional[fixieai.FixieClient] = None

    @property
    def client(self) -> fixieai.FixieClient:
        if self._client is None:
            if constants.is_authenticated():
                self._client = fixieai.FixieClient()
            else:
                click.secho(
                    "User is not authenticated. Run 'fixie auth' to authenticate, or "
                    "set the FIXIE_API_KEY environment variable, which can be obtained"
                    " from your profile page at https://app.fixie.ai/profile",
                    fg="red",
                )
                raise click.exceptions.Exit(11)
        return self._client


@click.group()
@click.pass_context
def fixie(ctx):
    """Command-line interface to the Fixie platform."""
    ctx.ensure_object(CliContext)


# Add subcommands
fixie.add_command(agent_commands.agent)
fixie.add_command(auth_commands.auth)
fixie.add_command(session_commands.session)
fixie.add_command(auth_commands.user)

# Add aliases for commonly used paths
fixie.add_command(agent_commands.init_agent, "init")
fixie.add_command(agent_commands.deploy, "deploy")
fixie.add_command(agent_commands.serve, "serve")
fixie.add_command(agent_commands.publish, "publish")
fixie.add_command(agent_commands.unpublish, "unpublish")
fixie.add_command(session_commands.new_session, "console")
fixie.add_command(session_commands.query, "query")


def get_installed_version(package_name: str) -> Optional[str]:
    try:
        # Get the distribution object for the specified package
        distribution = pkg_resources.get_distribution(package_name)
        # Get the version of the installed package
        installed_version = distribution.version
        return str(installed_version)
    except pkg_resources.DistributionNotFound:
        # Handle the case where the package is not found
        return None


def check_for_update(package_name: str):
    installed_version = get_installed_version(package_name)
    if not installed_version:
        return
    try:
        response = requests.get(f"https://pypi.org/pypi/{package_name}/json", timeout=3)
        response.raise_for_status()
    except Exception as e:
        click.secho(f"An error occurred while checking for updates: {e}", fg="yellow")
        return

    package_info = response.json()
    latest_version = package_info["info"]["version"]
    if version.parse(installed_version) < version.parse(latest_version):
        click.secho(
            f"🦊 A newer version of {package_name} is available: {latest_version}",
            fg="green",
        )

        click.secho(f"You are currently using version {installed_version}", fg="yellow")
        click.secho(f"You can upgrade using:", fg="yellow")
        click.secho(f"   pip install --upgrade {package_name}\n\n", fg="blue")


if __name__ == "__main__":
    check_for_update("fixieai")
    fixie()



import pytest

from fixieai.agents import utils


class ConversationalPrompt:
    """A wrapper type to facilitate testing prompts with/without the conversational flag."""

    def __init__(self, prompt: str):
        self.prompt = prompt

    def __str__(self):
        return self.prompt


BAD_PROMPTS = [
    # Ask Agent without Agent says
    """Q: Who is the president?
Ask Agent[search]: Who is the president?
A: It's Joe Biden.""",
    # Doesn't start with a query
    """Ask Agent[search]: Who is the president?
Q: Who is the president?
A: It's Joe Biden.""",
    # Ask Func without Func says
    """Q: How many books are there in the library?
Ask Func[library]: How many books are there in the library?
A: I don't know.""",
    # Doesn't end with a response
    """Q: What is 12 + 15?
Agent[calc] says: 27""",
    # 2 queries
    """Q: What is the life expectancy in the US?
Q: Do we need to call an agent?
A: No we don't.""",
    # 2 responses
    """Q: What is the life expectancy in the US?
A: I think it's 80 years but I'm not sure.
A: I'm not sure.""",
    # extra whitespace
    """Q: What is the life expectancy in the US?
A: I think it's 80 years but I'm not sure. """,
    # empty line at the end
    """Q: What is the life expectancy in the US?
A: I think it's 80 years but I'm not sure.
""",
    # empty line to start
    """
Q: What is the life expectancy in the US?
    A: I think it's 80 years but I'm not sure.""",
    # Invalid func name
    """Q: How many books are there in the library?
Ask Func[invalid_library]: How many books are there in the library?
Func[invalid_library] says: Many
A: There are many books in the library.""",
    # Mismatched func names
    """Q: How many books are there in the library?
Ask Func[library]: How many books are there in the library?
Func[bibliothèque] says: Many
A: There are many books in the library.""",
    # Mismatched agent names
    """Q: How many books are there in the library?
Ask Agent[library]: How many books are there in the library?
Agent[bibliothèque] says: Many
A: There are many books in the library.""",
    # Conversational without setting the flag
    """Q: Think of your favorite color
A: Okay, I'm thinking of it.
Q: Is it red?
A: No, it's not red.
Q: Is it blue?
A: Yes, it's blue!""",
    # Ask Agent lines aren't allowed between questions
    ConversationalPrompt(
        """Q: Think of your favorite color
A: Okay, I'm thinking of it.
Q: Is it red?
A: No, it's not red.
Ask Agent[library]: How many books are in the library?
Agent[library] says: I shouldn't be invoked between questions.
Q: Is it blue?
A: Yes, it's blue!"""
    ),
    # Ask Agent lines can't introduce new embeds
    """Q: What's the weather?
Ask Agent[weather]: What's the weather in #image1
Agent[weather] says: cloudy.
A: It's cloudy.""",
    # Ask Func lines can't introduce new embeds
    """Q: What's the weather?
Ask Func[weather]: #image1
Func[weather] says: cloudy.
A: It's cloudy.""",
    # Response lines can't introduce new embeds
    """Q: What's the weather?
A: Here's a picture of the weather #image1""",
    # Query lines must have a query
    """Q:
A: That's not a question!""",
]


@pytest.fixture
def is_valid_func_name():
    return lambda name: not name.startswith("invalid")


@pytest.mark.parametrize("bad_prompt", BAD_PROMPTS)
def test_fewshot_prompt_fails_with_bad_prompt(bad_prompt, is_valid_func_name):
    with pytest.raises(ValueError):
        utils._validate_few_shot_prompt(
            str(bad_prompt),
            conversational=isinstance(bad_prompt, ConversationalPrompt),
            is_valid_func_name=is_valid_func_name,
        )


GOOD_PROMPTS = [
    """Q: Who is the president?
Ask Agent[search]: Who is the president?
Agent[search] says: It's Joe Biden.
A: It's Joe Biden.""",
    """Q: How many books are there in the library?
Ask Func[library]: How many books are there in the library?
Func[library] says: None
A: I don't know.""",
    """Q: How many books are there in the library?
Func[library] says: Many
A: There are many books in the library""",
    """Q: What is 12 + 15?
Agent[calc] says: 27
A: It's 27.""",
    """Q: What is the life expectancy in the US?
A: I think it's 80 years but I'm not sure.""",
    """Q: Replace the background in #image1 with a beautiful sunset
Ask Agent[mask2former]: Mask out the background in #image1
Agent[mask2former] says: I have masked out the selected region: #mask1
Ask Func[edit]: A beautiful sunset #image1 #mask1
Func[edit] says: Here you go: #image2
A: I replaced the background with a beautiful sunset: #image2""",
    ConversationalPrompt(
        """Q: Think of your favorite color
A: Okay, I'm thinking of it.
Q: Is it red?
A: No, it's not red.
Q: Is it blue?
A: Yes, it's blue!"""
    ),
    """Q: Who is tweeting with the fixie hashtag?
Ask Func[hashtag_search]: ##fixie
Func[hashtag_search] says: @fixie
A: @fixie is currently tweeting with the fixie hashtag""",
]


@pytest.mark.parametrize("good_prompt", GOOD_PROMPTS)
def test_fewshot_prompt_succeeds_with_good_prompt(good_prompt, is_valid_func_name):
    utils._validate_few_shot_prompt(
        str(good_prompt),
        conversational=isinstance(good_prompt, ConversationalPrompt),
        is_valid_func_name=is_valid_func_name,
    )



"""This file holds some configuration constants, including URL to Fixie services,
and your Fixie API key."""

import os

# Base Fixie platform URL.
FIXIE_API_URL = os.getenv("FIXIE_API_URL", "https://app.fixie.ai")
# Path to fixie config file.
FIXIE_CONFIG_PATH = os.getenv(
    "FIXIE_CONFIG_PATH",
    os.path.expanduser("~/.config/fixie/config.yaml"),
)

# Fixie's GraphQL URL.
FIXIE_GRAPHQL_URL = f"{FIXIE_API_URL}/graphql"
# Fixie's UserStorage service URL.
FIXIE_USER_STORAGE_URL = f"{FIXIE_API_URL}/api/userstorage"
# Fixie's refresh endpoint. It will be pinged when your agent comes alive.
FIXIE_REFRESH_URL = f"{FIXIE_API_URL}/api/refresh"
# Fixie's OAuth redirect endpoint. Tokens sent to this endpoint will be redirected back
# to your agent.
FIXIE_OAUTH_REDIRECT_URL = f"{FIXIE_API_URL}/oauth"
# Fixie's deployments service URL.
FIXIE_DEPLOYMENT_URL = f"{FIXIE_API_URL}/api/deployments"
# Fixie's JWKS URL.
FIXIE_JWKS_URL = f"{FIXIE_API_URL}/.well-known/jwks.json"
# Fixie's OpenAI Proxy URL.
FIXIE_OPENAI_PROXY_URL = f"{FIXIE_API_URL}/api/openai-proxy/v1"
# Valid audiences for Fixie's query JWTs.
FIXIE_AGENT_API_AUDIENCES = ["https://app.fixie.ai/api", "https://app.dev.fixie.ai/api"]
# Fixie Agents' URL
FIXIE_AGENT_URL = f"{FIXIE_API_URL}/agents"


def fixie_api_key() -> str:
    """Returns authenticated user's fixie auth token.

    User may authenticate via `fixie auth`, or by setting FIXIE_API_KEY environment
    variable to override any previous authentication.

    If user is not authenticated and FIXIE_API_KEY is not set, a PermissionError is
    raised.
    """
    if "FIXIE_API_KEY" in os.environ:
        return os.environ["FIXIE_API_KEY"]
    try:
        from fixieai.cli.auth import user_config

        auth_token = user_config.load_config().auth_token
        assert isinstance(auth_token, str)
        return auth_token
    except (FileNotFoundError, KeyError, AssertionError):
        raise PermissionError(
            "User is not authenticated. Run 'fixie auth' to authenticate, or set the "
            "FIXIE_API_KEY environment variable, which can be obtained from your "
            "profile page at https://app.fixie.ai/profile"
        )


def is_authenticated() -> bool:
    """Returns true if the current user is authenticated."""
    try:
        fixie_api_key()
    except PermissionError:
        return False
    else:
        return True



from typing import Generator, Iterable, List, Optional

import pytest

import fixieai
from fixieai.agents import agent_func
from fixieai.agents import api
from fixieai.agents import corpora
from fixieai.agents import oauth
from fixieai.agents import token


@pytest.fixture
def oauth_params():
    return oauth.OAuthParams(
        client_id="dummy",
        client_secret="dummy",
        auth_uri="dummy",
        token_uri="dummy",
        scopes=["dummy"],
    )


def test_create_query_func_from_python_type_coercion(oauth_params):
    def valid(
        expected: Optional[Iterable[api.AgentResponse]] = None,
        oauth_params=None,
        default_message_type=str,
        allow_generator=False,
    ):
        def _wrap(f):
            normalized = agent_func.AgentQueryFunc.create(
                f, oauth_params, default_message_type, allow_generator
            )
            query_text = "Test query"
            result = normalized(
                api.AgentQuery(api.Message(text=query_text)),
                token.VerifiedTokenClaims("test-agent-id", False, "FAKE_TOKEN"),
            )

            expected_results = (
                [api.AgentResponse(api.Message(query_text))]
                if expected is None
                else list(expected)
            )
            assert expected_results == list(result)

        return _wrap

    @valid(expected=[api.AgentResponse(api.Message("PASS"))])
    def no_arguments():
        return "PASS"

    @valid()
    def untyped_query(query):
        return query

    @valid()
    def untyped_query_userstorage(query, user_storage):
        assert isinstance(user_storage, fixieai.UserStorage)
        return query

    @valid(oauth_params=oauth_params)
    def untyped_query_userstorage_oauth(query, user_storage, oauth_handler):
        assert isinstance(user_storage, fixieai.UserStorage)
        assert isinstance(oauth_handler, fixieai.OAuthHandler)
        return query

    @valid()
    def untyped_userstorage_query(user_storage, query):
        assert isinstance(user_storage, fixieai.UserStorage)
        return query

    @valid()
    def only_argument_query(query):
        return query

    @valid()
    def only_argument_assumed_to_be_query(x):
        return x

    @valid()
    def first_argument_assumed_to_be_query(x, user_storage):
        assert isinstance(user_storage, fixieai.UserStorage)
        return x

    @valid()
    def query_valid_type_annotation_str(x: str):
        assert isinstance(x, str)
        return x

    @valid()
    def query_valid_type_annotation_message(x: fixieai.Message):
        assert isinstance(x, fixieai.Message)
        return x.text

    @valid()
    def query_valid_type_annotation_query(x: fixieai.AgentQuery):
        assert isinstance(x, fixieai.AgentQuery)
        return x.message.text

    @valid()
    def valid_return_type_str(query: str) -> str:
        return query

    @valid()
    def valid_return_type_message(query: api.Message) -> api.Message:
        assert isinstance(query, api.Message)
        return query

    @valid()
    def valid_return_type_response(query: api.Message) -> api.AgentResponse:
        assert isinstance(query, api.Message)
        return api.AgentResponse(query)

    @valid(
        [api.AgentResponse(api.Message("1")), api.AgentResponse(api.Message("2"))],
        allow_generator=True,
    )
    def valid_return_type_response_generator(
        query: api.Message,
    ) -> api.AgentResponseGenerator:
        yield api.AgentResponse(api.Message("1"))
        yield api.AgentResponse(api.Message("2"))

    @valid(
        [api.AgentResponse(api.Message("1")), api.AgentResponse(api.Message("2"))],
        allow_generator=True,
    )
    def valid_return_type_str_generator(query) -> Generator:
        yield "1"
        yield "2"

    @valid(
        [api.AgentResponse(api.Message("1")), api.AgentResponse(api.Message("2"))],
        allow_generator=True,
    )
    def valid_return_type_str_generator_generic(query) -> Generator[str, None, None]:
        yield "1"
        yield "2"

    @valid(
        [api.AgentResponse(api.Message("1")), api.AgentResponse(api.Message("2"))],
        allow_generator=True,
    )
    def valid_return_type_str_iterable_generic(query) -> Iterable[str]:
        return ["1", "2"]

    @valid(
        [api.AgentResponse(api.Message("1")), api.AgentResponse(api.Message("2"))],
        allow_generator=True,
    )
    def valid_return_type_str_list_generic(query) -> List[str]:
        return ["1", "2"]

    @valid()
    def types_take_precedence(query: fixieai.UserStorage, user_storage: str):
        assert isinstance(query, fixieai.UserStorage)
        assert isinstance(user_storage, str)
        return user_storage

    @valid(default_message_type=str)
    def default_type_str(query):
        assert isinstance(query, str)
        return query

    @valid(default_message_type=fixieai.Message)
    def default_type_message(query):
        assert isinstance(query, fixieai.Message)
        return query

    @valid(default_message_type=fixieai.AgentQuery)
    def default_type_query(query):
        assert isinstance(query, fixieai.AgentQuery)
        return query.message

    def invalid(
        exception_type,
        match: str,
        oauth_params=None,
        default_message_type=str,
        allow_generator=False,
    ):
        def _wrap(f):
            with pytest.raises(exception_type, match=match):
                agent_func.AgentQueryFunc.create(
                    f, oauth_params, default_message_type, allow_generator
                )

        return _wrap

    @invalid(TypeError, "unknown parameters")
    def extra_argument_is_invalid(query, x):
        ...

    @invalid(TypeError, "unknown parameters")
    def too_many_arguments(a, b, c):
        ...

    @invalid(TypeError, "query")
    def query_invalid_type_annotation(query: int):
        ...

    @invalid(TypeError, "oauth_params")
    def oauth_handler_requires_oauth_params(query, user_storage, oauth_handler):
        ...

    @invalid(TypeError, "variable args")
    def varargs_not_allowed(*query):
        ...

    @invalid(TypeError, "variable args")
    def kwargs_not_allowed(**query):
        ...

    @invalid(TypeError, "returns", allow_generator=False)
    def generator_not_always_allowed(query: str) -> api.AgentResponseGenerator:
        yield api.AgentResponse(api.Message(query))

    @invalid(TypeError, "returns", allow_generator=True)
    def generator_invalid_generic() -> Generator[int, None, None]:
        yield 1


def test_create_corpus_func_from_python_type_coercion():
    REQUEST = corpora.CorpusRequest("part1", "pageToken")
    RESPONSE = corpora.CorpusResponse(
        [
            corpora.CorpusPartition("part1"),
            corpora.CorpusPartition("part2", "page0Token"),
        ],
        corpora.CorpusPage(
            [corpora.CorpusDocument("doc1", "doc1Contents".encode())], "token"
        ),
    )

    def valid():
        def _wrap(f):
            normalized = agent_func.AgentCorpusFunc.create(f)
            result = normalized(
                REQUEST,
                token.VerifiedTokenClaims("test-agent-id", False, "FAKE_TOKEN"),
            )
            assert [RESPONSE] == list(result)

        return _wrap

    @valid()
    def valid_no_arguments():
        return RESPONSE

    @valid()
    def valid_untyped_query(query):
        assert REQUEST == query
        return RESPONSE

    @valid()
    def valid_untyped_request(request):
        assert REQUEST == request
        return RESPONSE

    @valid()
    def valid_untyped_gibberish(gibberish):
        assert REQUEST == gibberish
        return RESPONSE

    @valid()
    def valid_type_annotation(x: corpora.CorpusRequest):
        assert REQUEST == x
        return RESPONSE

    def invalid(match: str):
        def _wrap(f):
            with pytest.raises(TypeError, match=match):
                agent_func.AgentCorpusFunc.create(f)

        return _wrap

    @invalid("must have one or zero arguments")
    def extra_argument_is_invalid(request, x):
        ...

    @invalid("query")
    def invalid_type_annotation(query: int):
        ...

    @invalid("variable args")
    def varargs_not_allowed(*query):
        ...

    @invalid("variable args")
    def kwargs_not_allowed(**query):
        ...

    @invalid("returns")
    def invalid_response_type(query) -> api.AgentResponse:
        return api.AgentResponse(api.Message(query))

    @invalid("returns")
    def void_response_type(query) -> None:
        return

    @invalid("returns")
    def invalid_response_type_generator() -> (
        Generator[corpora.CorpusResponse, None, None]
    ):
        yield RESPONSE



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fixieai',
 'fixieai.agents',
 'fixieai.cli',
 'fixieai.cli.agent',
 'fixieai.cli.auth',
 'fixieai.cli.session',
 'fixieai.client']

package_data = \
{'': ['*']}

install_requires = \
['Deprecated>=1.2.13,<2.0.0',
 'Pillow',
 'PyJWT[crypto]>=2.6.0,<3.0.0',
 'click>=8.1.3,<9.0.0',
 'dataclasses-json>=0.5.7,<0.6.0',
 'fastapi>=0.89.1,<0.90.0',
 'gql[requests]>=3.4.1,<4.0.0',
 'oauth2-client>=1.3.0,<2.0.0',
 'packaging>=23.0,<24.0',
 'prompt-toolkit',
 'pydantic>=1.10.8',
 'python-dotenv>=1.0.0,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'rich>=12.6.0,<13.0.0',
 'types-Deprecated>=1.2.9.2,<2.0.0.0',
 'uvicorn[standard]>=0.20.0,<0.21.0',
 'validators>=0.20.0,<0.21.0',
 'watchfiles>=0.19.0,<0.20.0',
 'wrapt>=1.15.0,<2.0.0']

entry_points = \
{'console_scripts': ['fixie = fixieai.cli.cli:fixie',
                     'fixieai = fixieai.cli.cli:fixie']}

setup_kwargs = {
    'name': 'fixieai',
    'version': '1.0.0',
    'description': 'SDK for the Fixie.ai platform. See: https://fixie.ai',
    'long_description': 'None',
    'author': 'Fixie.ai Team',
    'author_email': 'hello@fixie.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
